# SPDX-License-Identifier: MIT
"""Timing and data handling application wrapper for road events."""

import pygtk

pygtk.require("2.0")

import gtk
import glib
import gobject

import os
import sys
import logging
import random

import metarace

from metarace import ucsv
from metarace import jsonconfig
from metarace import tod
from metarace import riderdb
from metarace import eventdb
from metarace import telegraph
from metarace import export
from metarace import decoder
from metarace import timy
from metarace import strops
from metarace import loghandler
from metarace import namebank
from metarace import report
from metarace import uiutil
from metarace import irtt
from metarace import trtt
from metarace import rms

LOGFILE = u'event.log'
LOGFILE_LEVEL = logging.DEBUG
CONFIGFILE = u'config.json'
ROADMEET_ID = u'roadmeet_3.0'  # configuration versioning
EXPORTPATH = u'export'
_log = logging.getLogger(u'metarace.roadmeet')
_log.setLevel(logging.DEBUG)
ROADRACE_TYPES = {
    u'road': u'Road Race',
    u'circuit': u'Circuit',
    u'criterium': u'Criterium',
    u'handicap': u'Handicap',
    u'cross': u'Cyclocross',
    u'irtt': u'Road Time Trial',
    u'trtt': u'Team Road Time Trial',
}


class registerdlg(object):

    def __init__(self, meet=None):
        self.rdb = meet.rdb
        b = gtk.Builder()
        b.add_from_file(os.path.join(metarace.UI_PATH, u'transponder_reg.ui'))
        self.dlg = b.get_object(u'tagregister')
        self.dlg.set_transient_for(meet.window)
        self.rfidval = b.get_object(u'rfid_val')
        self.bibent = b.get_object(u'bib_entry')
        self.serent = b.get_object(u'series_entry')
        self.riderval = b.get_object(u'rider_val')
        self.autoinc = b.get_object(u'autoinc_but')
        self.autotrack = b.get_object(u'autotrack_but')
        self.storeid = None  # TODO: track scanned rfid with changing bib
        b.connect_signals(self)

    def series_entry_changed_cb(self, entry, data=None):
        bib = self.bibent.get_text().decode(u'utf-8', u'replace')
        if bib:
            self.bib_entry_changed_cb(self.bibent)

    def bib_entry_changed_cb(self, entry, data=None):
        bib = entry.get_text().decode(u'utf-8', u'replace')
        ser = self.serent.get_text().decode(u'utf-8', u'replace')
        r = self.rdb.getrider(bib, ser)
        if r is not None:
            first = self.rdb.getvalue(r, riderdb.COL_FIRST)
            last = self.rdb.getvalue(r, riderdb.COL_LAST)
            club = self.rdb.getvalue(r, riderdb.COL_ORG)
            cat = self.rdb.getvalue(r, riderdb.COL_CAT)
            refid = self.rdb.getvalue(r, riderdb.COL_REFID)
            self.riderval.set_text(u'{0} {1} ({2}) / {3}'.format(
                first, last, club, cat))
            if refid:
                self.rfidval.set_text(refid)
            else:
                self.rfidval.set_text(u'')
        else:
            self.riderval.set_text(u'[new rider]')
            self.rfidval.set_text(u'')

    def rfid_val_activate_cb(self, entry, data=None):
        """Activate on rfid val updates rider record."""
        bib = self.bibent.get_text().decode(u'utf-8', u'replace')
        ser = self.serent.get_text().decode(u'utf-8', u'replace')
        nrid = self.rfidval.get_text().decode(u'utf-8', u'replace')
        r = self.rdb.getrider(bib, ser)
        if r is not None:
            self.rdb.editrider(r, refid=nrid)
            _log.info(u'Updated transponder id for rider %r to %r', bib, nrid)

    def bib_entry_activate_cb(self, entry, data=None):
        """Activate on bib adds new rider, unless it exists."""
        bib = entry.get_text().decode(u'utf-8', u'replace')
        ser = self.serent.get_text().decode(u'utf-8', u'replace')
        r = self.rdb.getrider(bib, ser)
        if r is None:
            self.rdb.addempty(bib, ser)
            _log.debug(u'Added new rider %r:%r', bib, ser)

    def run(self):
        return self.dlg.run()

    def destroy(self):
        self.dlg.destroy()

    def increment_rider(self):
        cbib = self.bibent.get_text().decode(u'utf-8', u'replace').strip()
        cser = self.serent.get_text().decode(u'utf-8', u'replace').strip()
        _log.debug(u'Increment rider from %r:%r', cbib, cser)
        if cbib.isdigit():
            if self.autotrack.get_active():
                nbib = self.rdb.nextriderin(cbib, cser)
                _log.debug(u'Next rider from DB: %r:%r', nbib, cser)
                if nbib is not None:
                    self.bibent.set_text(nbib)
                    self.bibent.activate()
            else:
                nbib = int(cbib) + 1
                _log.debug(u'Next rider number: %r:%r', nbib, cser)
                self.bibent.set_text(unicode(nbib))
                self.bibent.activate()
        return False

    def register_tag(self, e):
        if e.refid:
            r = self.rdb.getrefid(e.refid)
            if r is not None:
                # transponder aleady assigned to a rider, load them
                bib = self.rdb.getvalue(r, riderdb.COL_BIB)
                ser = self.rdb.getvalue(r, riderdb.COL_SERIES)
                self.rfidval.set_text(e.refid)
                self.bibent.set_text(bib)
                self.serent.set_text(ser)
            else:
                # transponder currently unassigned
                bib = self.bibent.get_text().decode(u'utf-8', u'replace')
                if bib:
                    ser = self.serent.get_text().decode(u'utf-8', u'replace')
                    self.bibent.activate()  # required?
                    r = self.rdb.getrider(bib, ser)
                    if r is not None:
                        # check for existing tag allocation
                        orefid = self.rdb.getvalue(r, riderdb.COL_REFID)
                        if not orefid:
                            nrefid = e.refid.lower()
                            self.rdb.editrider(r, refid=nrefid)
                            _log.warning(u'Assigned ID %r to %r:%r', nrefid,
                                         bib, ser)
                            self.rfidval.set_text(nrefid)
                            if self.autoinc.get_active():
                                glib.idle_add(self.increment_rider)
                            else:
                                self.bibent.grab_focus()
                        else:
                            _log.warning(u'Rider %r:%r already assigned to %r',
                                         bib, ser, orefid)
                else:
                    _log.warning(u'Rider number entry empty')
                    self.rfidval.set_text(e.refid.lower())


class fakemeet(object):
    """Road meet placeholder for external event manipulations."""

    def __init__(self, edb, rdb):
        self.edb = edb
        self.rdb = rdb
        self.timer = decoder.decoder()
        self.alttimer = timy.timy()
        self.stat_but = uiutil.statbut(gtk.Button())
        self.action_model = gtk.ListStore(
            gobject.TYPE_STRING,
            gobject.TYPE_STRING,
        )
        self.action_model.append(['a', 'a'])
        self.action_combo = gtk.ComboBox()
        self.action_combo.set_model(self.action_model)
        self.action_combo.set_active(0)

        self.announce = telegraph.telegraph()
        self.title_str = ''
        self.host_str = ''
        self.subtitle_str = ''
        self.date_str = ''
        self.organiser_str = ''
        self.commissaire_str = ''
        self.distance = None
        self.docindex = 0
        self.linkbase = u'.'
        self.provisionalstart = False
        self.indexlink = None
        self.nextlink = None
        self.prevlink = None
        self.bibs_in_results = True

    def get_distance(self):
        return self.distance

    def cmd_announce(self, command, msg):
        return False

    def rider_announce(self, rvec):
        return False

    def timer_announce(self, evt, timer=None, source=u''):
        return False

    def report_strings(self, rep):
        """Copy the meet strings into the supplied report."""

        ## this is a copy of meet.print_report()
        rep.strings[u'title'] = self.title_str
        rep.strings[u'subtitle'] = self.subtitle_str
        rep.strings[u'host'] = self.host_str
        rep.strings[u'docstr'] = self.document_str
        rep.strings[u'datestr'] = strops.promptstr(u'Date:', self.date_str)
        rep.strings[u'commstr'] = strops.promptstr(u'Chief Commissaire:',
                                                   self.commissaire_str)
        rep.strings[u'orgstr'] = strops.promptstr(u'Organiser:',
                                                  self.organiser_str)
        if self.distance:
            rep.strings[u'diststr'] = strops.promptstr(
                u'Distance:',
                unicode(self.distance) + u'\u2006km')
        else:
            rep.strings[u'diststr'] = self.diststr

        if self.eventcode:
            rep.eventid = self.eventcode
        if self.prevlink:
            rep.prevlink = self.prevlink
        if self.nextlink:
            rep.nextlink = self.nextlink
        if self.indexlink:
            rep.indexlink = self.indexlink
        if self.shortname:
            rep.shortname = self.shortname

    def loadconfig(self):
        """Load meet config from disk."""
        cr = jsonconfig.config({
            u'roadmeet': {
                u'title': u'',
                u'shortname': u'',
                u'subtitle': u'',
                u'host': u'',
                u'document': u'',
                u'date': u'',
                u'organiser': u'',
                u'commissaire': u'',
                u'distance': None,
                u'diststr': u'',
                u'docindex': u'0',
                u'linkbase': u'.',
                u'indexlink': None,
                u'nextlink': None,
                u'prevlink': None,
                u'resultnos': u'Yes',
                u'provisionalstart': False,
                u'competitioncode': u'',
                u'eventcode': u'',
                u'racetype': u'',
                u'competitortype': u'',
                u'documentversion': u'',
                u'id': u''
            }
        })
        cr.add_section(u'roadmeet')
        cr.merge(metarace.sysconf, u'roadmeet')
        # check for config file
        try:
            with open(CONFIGFILE, 'rb') as f:
                cr.read(f)
        except Exception as e:
            _log.error(u'Error reading meet config: %s', e)
        # set meet meta, and then copy into text entries
        self.shortname = cr.get(u'roadmeet', u'shortname')
        self.title_str = cr.get(u'roadmeet', u'title')
        self.host_str = cr.get(u'roadmeet', u'host')
        self.subtitle_str = cr.get(u'roadmeet', u'subtitle')
        self.document_str = cr.get(u'roadmeet', u'document')
        self.date_str = cr.get(u'roadmeet', u'date')
        self.organiser_str = cr.get(u'roadmeet', u'organiser')
        self.commissaire_str = cr.get(u'roadmeet', u'commissaire')
        self.linkbase = cr.get(u'roadmeet', u'linkbase')
        self.distance = cr.get_float(u'roadmeet', u'distance')
        self.diststr = cr.get(u'roadmeet', u'diststr')
        self.docindex = cr.get_posint(u'roadmeet', u'docindex', 0)
        self.competitioncode = cr.get(u'roadmeet', u'competitioncode')
        self.eventcode = cr.get(u'roadmeet', u'eventcode')
        self.racetype = cr.get(u'roadmeet', u'racetype')
        self.linkbase = cr.get(u'roadmeet', u'linkbase')
        self.indexlink = cr.get(u'roadmeet', u'indexlink')
        self.prevlink = cr.get(u'roadmeet', u'prevlink')
        self.nextlink = cr.get(u'roadmeet', u'nextlink')
        self.competitortype = cr.get(u'roadmeet', u'competitortype')
        self.documentversion = cr.get(u'roadmeet', u'documentversion')
        self.bibs_in_results = cr.get_bool(u'roadmeet', u'resultnos')
        self.provisionalstart = cr.get_bool(u'roadmeet', u'provisionalstart')

    def event_configfile(self, evno):
        """Return a config filename for the given event no."""
        return u'event_{}.json'.format(unicode(evno))


class roadmeet(object):
    """Road meet application class."""

    ## Meet Menu Callbacks
    def menu_meet_save_cb(self, menuitem, data=None):
        """Save current all meet data to config."""
        self.saveconfig()

    def get_short_name(self):
        """Return the <= 16 char shortname."""
        return self.shortname

    def cat_but_auto_clicked(self, but, entry, data=None):
        """Lookup cats and write them into the supplied entry."""
        entry.set_text(u' '.join(self.rdb.listcats()))

    def menu_race_properties_activate_cb(self, menuitem, data=None):
        """Edit race specific properties."""
        if self.curevent is not None:
            _log.debug(u'Editing race properties')
            self.curevent.edit_event_properties(self.window)

    def menu_meet_properties_cb(self, menuitem, data=None):
        """Edit meet properties."""
        b = gtk.Builder()
        b.add_from_file(os.path.join(metarace.UI_PATH, u'roadmeet_props.ui'))
        dlg = b.get_object(u'properties')
        dlg.set_transient_for(self.window)

        # setup the type entry
        tcombo = b.get_object(u'type_combo')
        tmodel = b.get_object(u'type_model')
        tlbl = self.etype
        dotype = False
        # correct empty type
        if self.etype == u'':
            self.etype = u'road'
        if self.etype in ROADRACE_TYPES:
            tlbl = ROADRACE_TYPES[self.etype]
            dotype = True
            cnt = 0
            for t in [
                    u'road', u'circuit', u'handicap', u'criterium', u'cross',
                    u'irtt', u'trtt'
            ]:
                tmodel.append([t, ROADRACE_TYPES[t]])
                if t == self.etype:
                    tcombo.set_active(cnt)
                cnt += 1
            tcombo.set_sensitive(True)
        else:
            _log.warning(u'Unknown event type %r', self.etype)
            tmodel.append([self.etype, tlbl])
            tcombo.set_active(0)
            tcombo.set_sensitive(False)

        # fetch event result categories
        ocats = []
        cat_ent = b.get_object(u'cat_entry')
        if self.curevent is not None:
            ocats = self.curevent.get_catlist()
            cat_ent.set_text(u' '.join(ocats))
            cba = b.get_object(u'cat_but_auto')
            cba.connect(u'clicked', self.cat_but_auto_clicked, cat_ent)

        # fill text entries
        t_ent = b.get_object(u'title_entry')
        t_ent.set_text(self.title_str)
        st_ent = b.get_object(u'subtitle_entry')
        st_ent.set_text(self.subtitle_str)
        doc_ent = b.get_object(u'document_entry')
        doc_ent.set_text(self.document_str)
        d_ent = b.get_object(u'date_entry')
        d_ent.set_text(self.date_str)
        o_ent = b.get_object(u'organiser_entry')
        o_ent.set_text(self.organiser_str)
        c_ent = b.get_object(u'commissaire_entry')
        c_ent.set_text(self.commissaire_str)
        di_ent = b.get_object(u'distance_entry')
        if self.distance is not None:
            di_ent.set_text(str(self.distance))
        dis_ent = b.get_object(u'diststr_entry')
        if self.diststr is not None:
            dis_ent.set_text(self.diststr)
        ate = b.get_object(u'announce_topic_entry')
        if self.anntopic is not None:
            ate.set_text(self.anntopic)
        tte = b.get_object(u'timing_topic_entry')
        if self.timertopic is not None:
            tte.set_text(self.timertopic)
        ren = b.get_object(u'remote_enable_check')
        ren.set_active(self.remote_enable)
        mte = b.get_object(u'timing_main_entry')
        mte.set_text(self.timer_port)
        alte = b.get_object('timing_alt_entry')
        alte.set_text(self.alttimer_port)
        response = dlg.run()
        if response == 1:  # id 1 set in glade for "Apply"
            _log.debug(u'Updating meet properties')
            self.title_str = t_ent.get_text().decode(u'utf-8')
            self.subtitle_str = st_ent.get_text().decode(u'utf-8')
            self.document_str = doc_ent.get_text().decode(u'utf-8')
            self.date_str = d_ent.get_text().decode(u'utf-8')
            self.organiser_str = o_ent.get_text().decode(u'utf-8')
            self.commissaire_str = c_ent.get_text().decode(u'utf-8')
            self.distance = strops.confopt_float(
                di_ent.get_text().decode(u'utf-8'))
            self.diststr = dis_ent.get_text().decode(u'utf-8')

            # 'announce' topic
            ntopic = ate.get_text().decode(u'utf-8')
            if ntopic != self.anntopic:
                if self.anntopic is not None:
                    self.announce.unsubscribe(u'/'.join(
                        (self.anntopic, u'control', u'#')))
                self.anntopic = None
                if ntopic:
                    self.anntopic = ntopic
                    self.announce.subscribe(u'/'.join(
                        (self.anntopic, u'control', u'#')))
            # remote timer topic
            ntopic = tte.get_text().decode(u'utf-8')
            if ntopic != self.timertopic:
                if self.timertopic is not None:
                    self.announce.unsubscribe(self.timertopic)
                self.timertopic = None
                if ntopic:
                    self.timertopic = ntopic

            # update remote subscription
            self.remote_enable = ren.get_active()
            self.remote_reset()

            # reset timer
            self.set_timer(mte.get_text())

            nport = alte.get_text().decode(u'utf-8')
            if nport != self.alttimer_port:
                self.alttimer_port = nport
                self.alttimer.setport(nport)

            reload = False
            if self.curevent is not None:
                ncats = cat_ent.get_text().decode(u'utf-8').split()
                if ncats != ocats:
                    _log.debug(u'Result cats changed %r -> %r', ocats, ncats)
                    self.curevent.loadcats(ncats)
                    reload = True
            nt = tmodel.get_value(tcombo.get_active_iter(), 0).decode(u'utf-8')
            if dotype:
                # check for type change
                if nt != self.etype:
                    _log.info(u'Event type changed from %r to %r', self.etype,
                              nt)
                    if nt == u'crit':
                        self.curevent.downtimes(False)
                    else:
                        self.curevent.downtimes(True)
                    reload = True
            if reload:
                event = self.edb.getfirst()
                event[u'type'] = nt
                self.etype = nt
                self.menu_race_run_activate_cb(None, None)

            self.set_title()
            _log.debug(u'Properties updated')
        else:
            _log.debug(u'Edit properties cancelled')
        dlg.destroy()

    def print_report(self, sections=[], provisional=False):
        """Print the pre-formatted sections in a standard report."""
        rep = report.report()
        rep.provisional = provisional
        rep.strings[u'title'] = self.title_str
        rep.strings[u'host'] = self.host_str
        rep.strings[u'subtitle'] = self.subtitle_str
        rep.strings[u'docstr'] = self.document_str
        rep.strings[u'datestr'] = strops.promptstr(u'Date:', self.date_str)
        rep.strings[u'commstr'] = strops.promptstr(u'Chief Commissaire:',
                                                   self.commissaire_str)
        rep.strings[u'orgstr'] = strops.promptstr(u'Organiser:',
                                                  self.organiser_str)
        if self.distance:
            rep.strings[u'diststr'] = strops.promptstr(
                u'Distance:',
                unicode(self.distance) + u'\u2006km')
        else:
            rep.strings[u'diststr'] = self.diststr

        if self.eventcode:
            rep.eventid = self.eventcode
        if self.prevlink:
            rep.prevlink = self.prevlink
        if self.nextlink:
            rep.nextlink = self.nextlink
        if self.indexlink:
            rep.indexlink = self.indexlink
        if self.shortname:
            rep.shortname = self.shortname
        for sec in sections:
            rep.add_section(sec)

        print_op = gtk.PrintOperation()
        print_op.set_allow_async(True)
        print_op.set_print_settings(self.printprefs)
        print_op.set_default_page_setup(self.pageset)
        print_op.connect(u'begin_print', self.begin_print, rep)
        print_op.connect(u'draw_page', self.draw_print_page, rep)
        res = print_op.run(gtk.PRINT_OPERATION_ACTION_PREVIEW, None)
        if res == gtk.PRINT_OPERATION_RESULT_APPLY:
            self.printprefs = print_op.get_print_settings()
            _log.debug(u'Updated print preferences')
        elif res == gtk.PRINT_OPERATION_RESULT_IN_PROGRESS:
            _log.debug(u'Print operation in progress')
        self.docindex += 1

        # For convenience, also save copies to pdf and xls
        ofile = u'output.pdf'
        with metarace.savefile(ofile) as f:
            rep.output_pdf(f)
        ofile = u'output.xls'
        with metarace.savefile(ofile) as f:
            rep.output_xls(f)
        return False

    def begin_print(self, operation, context, rep):
        """Set print pages and units."""
        rep.start_gtkprint(context.get_cairo_context())
        operation.set_use_full_page(True)
        operation.set_n_pages(rep.get_pages())
        operation.set_unit('points')

    def draw_print_page(self, operation, context, page_nr, rep):
        """Draw to the nominated page."""
        rep.set_context(context.get_cairo_context())
        rep.draw_page(page_nr)

    def menu_meet_quit_cb(self, menuitem, data=None):
        """Quit the application."""
        self.running = False
        self.window.destroy()

    ## Race Menu Callbacks
    def menu_race_run_activate_cb(self, menuitem=None, data=None):
        """Open the event handler."""
        eh = self.edb.getfirst()
        if eh is not None:
            self.open_event(eh)
            self.set_title()

    def menu_race_close_activate_cb(self, menuitem, data=None):
        """Close callback - disabled in roadrace."""
        self.close_event()

    def menu_race_abort_activate_cb(self, menuitem, data=None):
        """Close the currently open event without saving."""
        if self.curevent is not None:
            self.curevent.readonly = True
        self.close_event()

    def menu_race_armstart_activate_cb(self, menuitem, data=None):
        """Default armstart handler."""
        _log.info(u'Arm Start')
        try:
            self.curevent.armstart()
        except Exception as e:
            _log.error(u'Arm start %s: %s', e.__class__.__name__, e)

    def menu_race_armlap_activate_cb(self, menuitem, data=None):
        """Default armlap handler."""
        _log.debug(u'Arm Lap')
        try:
            self.curevent.armlap()
        except Exception as e:
            _log.error(u'Arm lap %s: %s', e.__class__.__name__, e)

    def menu_race_armfin_activate_cb(self, menuitem, data=None):
        """Default armfin handler."""
        _log.info(u'Arm Finish')
        try:
            self.curevent.armfinish()
        except Exception as e:
            _log.error(u'Arm finish %s: %s', e.__class__.__name__, e)

    def menu_race_finished_activate_cb(self, menuitem, data=None):
        """Default finished handler."""
        _log.info(u'Finished')
        try:
            self.curevent.set_finished()
        except Exception as e:
            _log.error(u'Set finished %s: %s', e.__class__.__name__, e)

    def open_event(self, eventhdl=None):
        """Open provided event handle."""
        if eventhdl is not None:
            self.close_event()
            if self.etype not in ROADRACE_TYPES:
                _log.warning(u'Unknown event type %r', self.etype)
            if self.etype == u'irtt':
                self.curevent = irtt.irtt(self, eventhdl, True)
            elif self.etype == u'trtt':
                self.curevent = trtt.trtt(self, eventhdl, True)
            else:
                self.curevent = rms.rms(self, eventhdl, True)

            self.curevent.loadconfig()
            self.race_box.add(self.curevent.frame)

            # re-populate the rider command model.
            cmdo = self.curevent.get_ridercmdorder()
            cmds = self.curevent.get_ridercmds()
            if cmds is not None:
                self.action_model.clear()
                for cmd in cmdo:
                    self.action_model.append([cmd, cmds[cmd]])
                self.action_combo.set_active(0)

            self.menu_race_close.set_sensitive(True)
            self.menu_race_abort.set_sensitive(True)
            starters = eventhdl[u'star']
            if starters is not None and starters != u'':
                self.curevent.race_ctrl(u'add', starters)
                eventhdl[u'star'] = u''  # and clear
            self.curevent.show()

    def close_event(self):
        """Close the currently opened race."""
        if self.curevent is not None:
            self.curevent.hide()
            self.race_box.remove(self.curevent.frame)
            self.curevent.destroy()
            self.menu_race_close.set_sensitive(False)
            self.menu_race_abort.set_sensitive(False)
            self.curevent = None
            self.stat_but.buttonchg(uiutil.bg_none, u'Closed')
            self.stat_but.set_sensitive(False)

    ## Reports menu callbacks.
    def menu_reports_startlist_activate_cb(self, menuitem, data=None):
        """Generate a startlist."""
        if self.curevent is not None:
            sections = self.curevent.startlist_report()
            if sections:
                self.print_report(sections)
            else:
                _log.info(u'Startlist - Nothing to print')

    def menu_reports_callup_activate_cb(self, menuitem, data=None):
        """Generate a start line call-up."""
        if self.curevent is not None:
            sections = self.curevent.callup_report()
            if sections:
                self.print_report(sections)
            else:
                _log.info(u'Callup - Nothing to print')

    def menu_reports_signon_activate_cb(self, menuitem, data=None):
        """Generate a sign on sheet."""
        if self.curevent is not None:
            sections = self.curevent.signon_report()
            if sections:
                self.print_report(sections)
            else:
                _log.info(u'Sign-on - Nothing to print')

    def menu_reports_analysis_activate_cb(self, menuitem, data=None):
        """Generate the analysis report."""
        if self.curevent is not None:
            sections = self.curevent.analysis_report()
            if sections:
                self.print_report(sections)

    def menu_reports_camera_activate_cb(self, menuitem, data=None):
        """Generate the camera operator report."""
        if self.curevent is not None:
            sections = self.curevent.camera_report()
            if sections:
                self.print_report(sections)

    def race_results_points_activate_cb(self, menuitem, data=None):
        """Generate the points tally report."""
        if self.curevent is not None:
            sections = self.curevent.points_report()
            if sections:
                self.print_report(sections)

    def menu_reports_result_activate_cb(self, menuitem, data=None):
        """Generate the race result report."""
        if self.curevent is not None:
            sections = self.curevent.result_report()
            if sections:
                self.print_report(sections,
                                  self.curevent.timerstat != u'finished')

    def menu_data_rego_activate_cb(self, menuitem, data=None):
        """Open transponder registration dialog."""
        rdlg = registerdlg(self)
        ocb = self.timercb
        _log.debug(u'Save ocb %r', ocb)
        try:
            self.timercb = rdlg.register_tag
            rdlg.run()
        finally:
            self.timercb = ocb
            _log.debug(u'Restored cb %r', ocb)
        rdlg.destroy()

    def menu_data_uscb_activate_cb(self, menuitem, data=None):
        """Reload rider db from disk."""
        self.rdb.clear()
        self.rdb.load(u'riders.csv')
        _log.info(u'Reloaded riders from disk')
        self.menu_race_run_activate_cb()

    def menu_import_riders_activate_cb(self, menuitem, data=None):
        """Add riders to database."""
        sfile = uiutil.loadcsvdlg(u'Select rider file to import', self.window,
                                  u'.')
        if sfile is not None:
            with namebank.namebank() as n:
                self.rdb.load(sfile, namedb=n, overwrite=True)
            _log.info(u'Import riders from %r', sfile)
        else:
            _log.debug(u'Import riders cancelled')

    def menu_import_chipfile_activate_cb(self, menuitem, data=None):
        """Import a transponder chipfile."""
        sfile = uiutil.loadcsvdlg(u'Select chipfile to import', self.window,
                                  u'.')
        if sfile is not None:
            self.rdb.load_chipfile(sfile)
            _log.info(u'Import chipfile %r', sfile)
        else:
            _log.debug(u'Import chipfile cancelled')

    def menu_import_startlist_activate_cb(self, menuitem, data=None):
        """Import a startlist."""
        if self.curevent is None:
            _log.info(u'No event open for starters import')
            return
        count = 0
        sfile = uiutil.loadcsvdlg(u'Select startlist file to import',
                                  self.window, u'.')
        if os.path.isfile(sfile):
            with open(sfile, 'rb') as f:
                cr = ucsv.UnicodeReader(f)
                for r in cr:
                    if len(r) > 1 and r[1].isalnum() and r[1].lower() not in [
                            'no', 'no.'
                    ]:
                        bib = r[1].strip().lower()
                        series = u''
                        if len(r) > 2:
                            series = r[2].strip()
                        self.curevent.addrider(bib, series)
                        start = tod.mktod(r[0])
                        if start is not None:
                            self.curevent.starttime(start, bib, series)
                        count += 1
            _log.info(u'Import %r starters from %r', count, sfile)
        else:
            _log.debug(u'Import startlist cancelled')

    def menu_export_riders_activate_cb(self, menuitem, data=None):
        """Export rider database."""
        sfile = uiutil.savecsvdlg(u'Select file to export riders to',
                                  self.window, u'riders_export.csv', u'.')
        if sfile is not None:
            self.rdb.save(sfile)
            _log.info(u'Export rider data to %r', sfile)
        else:
            _log.debug(u'Export rider data cancelled')

    def menu_export_chipfile_activate_cb(self, menuitem, data=None):
        """Export transponder chipfile from rider model."""
        sfile = uiutil.savecsvdlg(u'Select file to export refids to',
                                  self.window, u'chipfile.csv', u'.')
        if sfile is not None:
            self.rdb.save_chipfile(sfile)
            _log.info(u'Export chipfile to %r', sfile)
        else:
            _log.debug(u'Export chipfile cancelled')

    def menu_export_result_activate_cb(self, menuitem, data=None):
        """Export raw result to disk."""
        if self.curevent is None:
            _log.info(u'No event open')
            return

        rfilename = uiutil.savecsvdlg(u'Select file to save results to.',
                                      self.window, u'results.csv', u'.')
        if rfilename is not None:
            with metarace.savefile(rfilename) as f:
                cw = ucsv.UnicodeWriter(f)
                cw.writerow([u'Rank', u'No.', u'Time', u'Bonus', u'Penalty'])
                for r in self.curevent.result_gen(u''):
                    opr = [u'', u'', u'', u'', u'']
                    for i in range(0, 2):
                        if r[i]:
                            opr[i] = unicode(r[i])
                    for i in range(2, 5):
                        if r[i]:
                            opr[i] = unicode(r[i].timeval)
                    cw.writerow(opr)
            _log.info(u'Export result to %r', rfilename)

    def menu_export_startlist_activate_cb(self, menuitem, data=None):
        """Extract startlist from current event."""
        if self.curevent is None:
            _log.info(u'No event open')
            return

        rfilename = uiutil.savecsvdlg(u'Select file to save startlist to.',
                                      self.window, u'startlist.csv', u'.')
        if rfilename is not None:
            with metarace.savefile(rfilename) as f:
                cw = ucsv.UnicodeWriter(f)
                cw.writerow([u'Start', u'No.', u'Series', u'Name', u'Cat'])
                if self.etype == u'irtt':
                    for r in self.curevent.startlist_gen():
                        cw.writerow(r)
                else:
                    clist = self.curevent.get_catlist()
                    clist.append(u'')
                    for c in clist:
                        for r in self.curevent.startlist_gen(c):
                            cw.writerow(r)

            _log.info(u'Export startlist to %r', rfilename)

    def export_result_maker(self):
        if self.mirrorfile:
            filebase = self.mirrorfile
        else:
            filebase = u'.'
        if filebase in [u'', u'.']:
            filebase = u''
            _log.warn(u'Using default filenames for export')
        else:
            pass

        fnv = []
        if filebase:
            fnv.append(filebase)
        fnv.append(u'startlist')
        sfile = u'_'.join(fnv)
        fnv[-1] = 'result'
        ffile = u'_'.join(fnv)

        # Write out a startlist if event idle
        if self.curevent.timerstat in [u'idle'] or self.etype == u'irtt':
            filename = sfile
            rep = report.report()
            rep.strings[u'title'] = self.title_str
            rep.strings[u'host'] = self.host_str
            rep.strings[u'subtitle'] = self.subtitle_str
            rep.strings[u'docstr'] = self.document_str
            rep.strings[u'datestr'] = strops.promptstr(u'Date:', self.date_str)
            rep.strings[u'commstr'] = strops.promptstr(u'Chief Commissaire:',
                                                       self.commissaire_str)
            rep.strings[u'orgstr'] = strops.promptstr(u'Organiser:',
                                                      self.organiser_str)
            if self.distance:
                rep.strings[u'diststr'] = strops.promptstr(
                    u'Distance:',
                    unicode(self.distance) + u'\u2006km')
            else:
                rep.strings[u'diststr'] = self.diststr
            if self.provisionalstart:
                rep.set_provisional(True)
            rep.indexlink = u'index'
            if self.eventcode:
                rep.eventid = self.eventcode
            if self.prevlink:
                rep.prevlink = u'_'.join((self.prevlink, u'startlist'))
            if self.nextlink:
                rep.nextlink = u'_'.join((self.nextlink, u'startlist'))
            if self.indexlink:
                rep.indexlink = self.indexlink
            if self.shortname:
                rep.shortname = self.shortname
            rep.resultlink = ffile
            for sec in self.curevent.startlist_report():
                rep.add_section(sec)

            lb = os.path.join(self.linkbase, filename)
            lt = [u'pdf', u'xls']
            rep.canonical = u'.'.join([lb, u'json'])
            ofile = os.path.join(self.exportpath, filename + u'.pdf')
            with metarace.savefile(ofile) as f:
                rep.output_pdf(f)
            ofile = os.path.join(self.exportpath, filename + u'.xls')
            with metarace.savefile(ofile) as f:
                rep.output_xls(f)
            ofile = os.path.join(self.exportpath, filename + u'.json')
            with metarace.savefile(ofile) as f:
                rep.output_json(f)
            ofile = os.path.join(self.exportpath, filename + u'.html')
            with metarace.savefile(ofile) as f:
                rep.output_html(f, linkbase=lb, linktypes=lt)

        # Then export a result
        rep = report.report()
        rep.strings[u'title'] = self.title_str
        rep.strings[u'host'] = self.host_str
        rep.strings[u'subtitle'] = self.subtitle_str
        rep.strings[u'docstr'] = self.document_str
        rep.strings[u'datestr'] = strops.promptstr(u'Date:', self.date_str)
        rep.strings[u'commstr'] = strops.promptstr(u'Chief Commissaire:',
                                                   self.commissaire_str)
        rep.strings[u'orgstr'] = strops.promptstr(u'Organiser:',
                                                  self.organiser_str)
        if self.distance:
            rep.strings[u'diststr'] = strops.promptstr(
                u'Distance:',
                unicode(self.distance) + u'\u2006km')
        else:
            rep.strings[u'diststr'] = self.diststr

        # Set provisional status	# TODO: other tests for prov flag?
        if self.curevent.timerstat != u'finished':
            rep.set_provisional(True)
        else:
            rep.reportstatus = u'final'  # TODO: write in other phases
        for sec in self.curevent.result_report():
            rep.add_section(sec)

        filename = ffile
        rep.indexlink = u'index'
        if self.eventcode:
            rep.eventid = self.eventcode
        if self.prevlink:
            rep.prevlink = u'_'.join((self.prevlink, u'result'))
        if self.nextlink:
            rep.nextlink = u'_'.join((self.nextlink, u'result'))
        if self.indexlink:
            rep.indexlink = self.indexlink
        if self.shortname:
            rep.shortname = self.shortname
        rep.startlink = sfile
        lb = os.path.join(self.linkbase, filename)
        lt = [u'pdf', u'xls']
        rep.canonical = u'.'.join([lb, u'json'])

        ofile = os.path.join(self.exportpath, filename + u'.pdf')
        with metarace.savefile(ofile) as f:
            rep.output_pdf(f)
        ofile = os.path.join(self.exportpath, filename + u'.xls')
        with metarace.savefile(ofile) as f:
            rep.output_xls(f)
        ofile = os.path.join(self.exportpath, filename + u'.json')
        with metarace.savefile(ofile) as f:
            rep.output_json(f)
        ofile = os.path.join(self.exportpath, filename + u'.html')
        with metarace.savefile(ofile) as f:
            rep.output_html(f, linkbase=lb, linktypes=lt)

    def menu_data_results_cb(self, menuitem, data=None):
        """Create live result report and export"""
        self.saveconfig()
        if self.curevent is None:
            return
        if self.lifexport:  # save current lif with export
            lifdat = self.curevent.lifexport()
            if len(lifdat) > 0:
                liffile = os.path.join(self.exportpath, u'lifexport.lif')
                with metarace.savefile(liffile) as f:
                    cw = ucsv.UnicodeWriter(f, quoting=csv.QUOTE_MINIMAL)
                    for r in lifdat:
                        cw.writerow(r)
        if self.resfiles:
            self.export_result_maker()
        glib.idle_add(self.mirror_start)

    ## Directory utilities
    def event_configfile(self, evno):
        """Return a config filename for the given event no."""
        return u'event_{}.json'.format(unicode(evno))

    ## Timing menu callbacks
    def menu_timing_status_cb(self, menuitem, data=None):
        if self.timer_port:
            if self.timer.connected():
                _log.info(u'Request timer status')
                self.timer.status()
            else:
                _log.info(u'Decoder disconnected')
        else:
            _log.info(u'No decoder configured')
        # always call into alt timer
        self.alttimer.status()

    def menu_timing_subtract_activate_cb(self, menuitem, data=None):
        """Run the time of day subtraction dialog."""
        b = gtk.Builder()
        b.add_from_file(os.path.join(metarace.UI_PATH, u'tod_subtract.ui'))
        ste = b.get_object(u'timing_start_entry')
        fte = b.get_object(u'timing_finish_entry')
        nte = b.get_object(u'timing_net_entry')
        b.get_object(u'timing_start_now').connect(u'clicked',
                                                  self.entry_set_now, ste)
        b.get_object(u'timing_finish_now').connect(u'clicked',
                                                   self.entry_set_now, fte)
        ste.connect(u'activate', self.menu_timing_recalc, ste, fte, nte)
        fte.connect(u'activate', self.menu_timing_recalc, ste, fte, nte)
        dlg = b.get_object(u'timing')
        dlg.set_transient_for(self.window)
        dlg.run()
        dlg.destroy()

    def menu_timing_start_activate_cb(self, menuitem, data=None):
        """Manually set race elapsed time via trigger."""
        if self.curevent is None:
            _log.info(u'No event open to set elapsed time on')
        else:
            self.curevent.elapsed_dlg()

    def entry_set_now(self, button, entry=None):
        """Enter the current time in the provided entry."""
        entry.set_text(tod.now().timestr())
        entry.activate()

    def menu_timing_recalc(self, entry, ste, fte, nte):
        """Update the net time entry for the supplied start and finish."""
        st = tod.mktod(ste.get_text())
        ft = tod.mktod(fte.get_text())
        if st is not None and ft is not None:
            ste.set_text(st.timestr())
            fte.set_text(ft.timestr())
            nte.set_text((ft - st).timestr())

    def menu_timing_clear_activate_cb(self, menuitem, data=None):
        """Start a new timing session in attached timers"""
        # Note: clear will perform reset, stop_session, clear,
        # sync, and start_session in whatever order is appropriate
        # for the decoder type
        self.timer.clear()
        self.alttimer.clrmem()

    def menu_timing_reconnect_activate_cb(self, menuitem, data=None):
        """Drop current timer connection and re-connect"""
        self.set_timer(self.timer_port, force=True)
        self.alttimer.setport(self.alttimer_port)
        self.alttimer.sane()
        if self.etype == u'irtt':
            self.alttimer.write(u'DTS05.00')
            self.alttimer.write(u'DTF00.01')
        else:
            # assume 1 second gaps at finish
            self.alttimer.write(u'DTF01.00')
        _log.info(u'Re-connect/re-start attached timers')

    def restart_decoder(self, data=None):
        """Request re-start of decoder."""
        self.timer.start_session()
        return None

    def menu_timing_configure_activate_cb(self, menuitem, data=None):
        """Attempt to re-configure the attached decoder from saved config."""
        if self.timer.__class__.__name__ == u'thbc':
            if not self.timer.connected():
                _log.info(u'Timer not connected, config not possible')
                return False
            if not uiutil.questiondlg(
                    self.window, u'Re-configure THBC Decoder Settings?',
                    u'Note: Passings will not be captured while decoder is updating.'
            ):
                _log.debug(u'Config aborted')
                return False
            self.timer.stop_session()
            self.timer.sane()
            glib.timeout_add_seconds(60, self.restart_decoder)
            self.timer.ipconfig()
        else:
            _log.info(u'Decoder config not available')
        return None

    ## Help menu callbacks
    def menu_help_about_cb(self, menuitem, data=None):
        """Display metarace about dialog."""
        metarace.about_dlg(self.window)

    ## Race Control Elem callbacks
    def race_stat_but_clicked_cb(self, button, data=None):
        """Call through into event if open."""
        if self.curevent is not None:
            self.curevent.stat_but_clicked(button)

    def race_stat_entry_activate_cb(self, entry, data=None):
        """Pass the chosen action and bib list through to curevent."""
        action = self.action_model.get_value(
            self.action_combo.get_active_iter(), 0)
        if self.curevent is not None:
            if self.curevent.race_ctrl(
                    action,
                    self.action_entry.get_text().decode(u'utf-8')):
                self.action_entry.set_text(u'')

    ## Menu button callbacks
    def race_action_combo_changed_cb(self, combo, data=None):
        """Notify curevent of change in combo."""
        aiter = self.action_combo.get_active_iter()
        if self.curevent is not None and aiter is not None:
            action = self.action_model.get_value(aiter, 0)
            self.curevent.ctrl_change(action, self.action_entry)

    def menu_clock_clicked_cb(self, button, data=None):
        """Handle click on menubar clock."""
        _log.info(u'PC ToD: %s', tod.now().rawtime())

    ## 'Slow' Timer callback - this is the main ui event routine
    def timeout(self):
        """Update status buttons and time of day clock button."""
        if self.running:
            # call into race timeout handler
            if self.curevent is not None:
                self.curevent.timeout()

            # check for completion in the export thread
            if self.mirror is not None:
                if not self.mirror.is_alive():
                    self.mirror = None
                    _log.debug(u'Removing completed export thread.')

            # update the menu status button
            nt = tod.now().meridiem()
            if self.rfuact:
                self.rfustat.buttonchg(uiutil.bg_armint, nt)
            else:
                if self.timer_port:
                    if self.timer.connected():
                        self.rfustat.buttonchg(uiutil.bg_armstart, nt)
                    else:
                        self.rfustat.buttonchg(uiutil.bg_armfin, nt)
                else:
                    self.rfustat.buttonchg(uiutil.bg_none, nt)
            self.rfuact = False

            # attempt to heal a broken lonk
            if self.timer_port:
                if self.timer.connected():
                    self.rfufail = 0
                else:
                    self.rfufail += 1
                    if self.rfufail > 10:
                        self.rfufail = 0
                        eport = self.timer_port.split(u':', 1)[-1]
                        self.timer.setport(eport)
            else:
                self.rfufail = 0
        else:
            return False
        return True

    ## Window methods
    def set_title(self, extra=u''):
        """Update window title from meet properties."""
        tv = []
        if self.etype in ROADRACE_TYPES:
            tv.append(ROADRACE_TYPES[self.etype] + u':')

        title = self.title_str.strip()
        if title:
            tv.append(title)
        subtitle = self.subtitle_str.strip()
        if subtitle:
            tv.append(subtitle)
        self.window.set_title(u' '.join(tv))
        if self.curevent is not None:
            self.curevent.set_titlestr(subtitle)

    def meet_destroy_cb(self, window, msg=u''):
        """Handle destroy signal and exit application."""
        rootlogger = logging.getLogger()
        rootlogger.removeHandler(self.sh)
        rootlogger.removeHandler(self.lh)
        self.window.hide()
        glib.idle_add(self.meet_destroy_handler)

    def meet_destroy_handler(self):
        if self.curevent is not None:
            self.close_event()
        if self.started:
            self.saveconfig()
            self.shutdown()  # threads are joined in shutdown
        rootlogger = logging.getLogger()
        if self.loghandler is not None:
            rootlogger.removeHandler(self.loghandler)
        self.running = False
        gtk.main_quit()
        return False

    def key_event(self, widget, event):
        """Collect key events on main window and send to race."""
        if event.type == gtk.gdk.KEY_PRESS:
            key = gtk.gdk.keyval_name(event.keyval) or u'None'  # str
            if event.state & gtk.gdk.CONTROL_MASK:
                key = key.lower()
                t = tod.now(chan=u'MAN', refid=unicode(key))
                if key in ['0', '1']:
                    # trigger
                    t.refid = u''
                    t.chan = strops.id2chan(strops.chan2id(key))
                    self._alttimercb(t)
                    return True
                elif key in ['2', '3', '4', '5', '6', '7', '8', '9']:
                    # passing
                    self._timercb(t)
                    return True
            if self.curevent is not None:
                return self.curevent.key_event(widget, event)
        return False

    def shutdown(self, msg=u''):
        """Shutdown worker threads and close application."""
        self.started = False
        self.announce.exit(msg)
        self.timer.exit(msg)
        self.alttimer.exit(msg)
        _log.info(u'Waiting for workers')
        if self.mirror is not None:
            _log.debug(u'Result export')
            self.mirror.join()
            self.mirror = None
        _log.debug(u'Telegraph/announce')
        self.announce.join()

    def start(self):
        """Start the timer and rfu threads."""
        if not self.started:
            _log.debug(u'Meet startup')
            self.announce.start()
            self.timer.start()
            self.alttimer.start()
            self.started = True

    ## Roadmeet functions
    def saveconfig(self):
        """Save current meet data to disk."""
        if self.curevent is not None and self.curevent.winopen:
            self.curevent.saveconfig()
        cw = jsonconfig.config()
        cw.add_section(u'roadmeet')
        cw.set(u'roadmeet', u'id', ROADMEET_ID)
        cw.set(u'roadmeet', u'anntopic', self.anntopic)
        cw.set(u'roadmeet', u'timertopic', self.timertopic)
        cw.set(u'roadmeet', u'remote_enable', self.remote_enable)
        cw.set(u'roadmeet', u'timer', self.timer_port)
        cw.set(u'roadmeet', u'alttimer', self.alttimer_port)
        cw.set(u'roadmeet', u'shortname', self.shortname)
        cw.set(u'roadmeet', u'linkbase', self.linkbase)
        cw.set(u'roadmeet', u'indexlink', self.indexlink)
        cw.set(u'roadmeet', u'nextlink', self.nextlink)
        cw.set(u'roadmeet', u'prevlink', self.prevlink)
        cw.set(u'roadmeet', u'title', self.title_str)
        cw.set(u'roadmeet', u'host', self.host_str)
        cw.set(u'roadmeet', u'subtitle', self.subtitle_str)
        cw.set(u'roadmeet', u'document', self.document_str)
        cw.set(u'roadmeet', u'date', self.date_str)
        cw.set(u'roadmeet', u'organiser', self.organiser_str)
        cw.set(u'roadmeet', u'commissaire', self.commissaire_str)

        cw.set(u'roadmeet', u'resultnos', self.bibs_in_results)
        cw.set(u'roadmeet', u'lifexport', self.lifexport)
        cw.set(u'roadmeet', u'resfiles', self.resfiles)
        cw.set(u'roadmeet', u'provisionalstart', self.provisionalstart)
        cw.set(u'roadmeet', u'distance', self.distance)
        cw.set(u'roadmeet', u'diststr', self.diststr)
        cw.set(u'roadmeet', u'docindex', self.docindex)
        cw.set(u'roadmeet', u'mirrorpath', self.mirrorpath)
        cw.set(u'roadmeet', u'mirrorcmd', self.mirrorcmd)
        cw.set(u'roadmeet', u'mirrorfile', self.mirrorfile)
        cw.set(u'roadmeet', u'competitioncode', self.competitioncode)
        cw.set(u'roadmeet', u'eventcode', self.eventcode)
        cw.set(u'roadmeet', u'racetype', self.racetype)
        cw.set(u'roadmeet', u'competitortype', self.competitortype)
        cw.set(u'roadmeet', u'documentversion', self.documentversion)

        with metarace.savefile(CONFIGFILE) as f:
            cw.write(f)
        self.rdb.save(u'riders.csv')
        self.edb.save(u'events.csv')
        _log.info(u'Meet configuration saved')

    def set_timer(self, newdevice=u'', force=False):
        """Re-set the main timer device and connect callback."""
        if newdevice != self.timer_port or force:
            self.timer = decoder.mkdevice(newdevice, self.timer)
            self.timer_port = newdevice
        else:
            _log.debug(u'set_timer - No change required')
        self.timer.setcb(self._timercb)

    def loadconfig(self):
        """Load meet config from disk."""
        cr = jsonconfig.config({
            u'roadmeet': {
                u'shortname': None,
                u'title': u'',
                u'host': u'',
                u'subtitle': u'',
                u'document': u'',
                u'date': u'',
                u'organiser': u'',
                u'commissaire': u'',
                u'distance': None,
                u'diststr': u'',
                u'docindex': u'0',
                u'timer': u'',
                u'alttimer': u'',
                u'resultnos': True,
                u'anntopic': None,
                u'timertopic': None,
                u'remote_enable': False,
                u'linkbase': u'.',
                u'indexlink': None,
                u'nextlink': None,
                u'prevlink': None,
                u'lifexport': False,
                u'resfiles': True,
                u'provisionalstart': False,
                u'mirrorpath': u'',
                u'mirrorcmd': u'echo',
                u'mirrorfile': u'',
                u'competitioncode': u'',
                u'eventcode': u'',
                u'racetype': u'',
                u'competitortype': u'',
                u'documentversion': u'',
                u'id': u''
            }
        })
        cr.add_section(u'roadmeet')
        cr.merge(metarace.sysconf, u'roadmeet')
        _log.debug(u'Load system meet defaults')

        # re-set main log file
        _log.debug(u'Adding meet logfile handler %r', LOGFILE)
        rootlogger = logging.getLogger()
        if self.loghandler is not None:
            rootlogger.removeHandler(self.loghandler)
            self.loghandler.close()
            self.loghandler = None
        self.loghandler = logging.FileHandler(LOGFILE)
        self.loghandler.setLevel(LOGFILE_LEVEL)
        self.loghandler.setFormatter(logging.Formatter(metarace.LOGFILEFORMAT))
        rootlogger.addHandler(self.loghandler)

        # check for config file
        if os.path.exists(CONFIGFILE):
            try:
                with open(CONFIGFILE, 'rb') as f:
                    cr.read(f)
                _log.debug(u'Read meet config from %r', CONFIGFILE)
            except Exception as e:
                _log.error(u'Unable to read meet config: %s', e)

        # set timer port (decoder)
        self.set_timer(cr.get(u'roadmeet', u'timer'))

        # set alt timer port (timy)
        nport = cr.get(u'roadmeet', u'alttimer')
        if nport != self.alttimer_port:
            self.alttimer_port = nport
            self.alttimer.setport(nport)
            self.alttimer.sane()

        # set the default announce topic and subscribe to control topic
        self.anntopic = cr.get(u'roadmeet', u'anntopic')
        if self.anntopic:
            self.announce.subscribe(u'/'.join(
                (self.anntopic, u'control', u'#')))

        # fetch the remote timer topic and update remote control
        self.timertopic = cr.get(u'roadmeet', u'timertopic')
        self.remote_enable = cr.get_bool(u'roadmeet', u'remote_enable')
        self.remote_reset()

        # set meet meta, and then copy into text entries
        self.shortname = cr.get(u'roadmeet', u'shortname')
        self.title_str = cr.get(u'roadmeet', u'title')
        self.host_str = cr.get(u'roadmeet', u'host')
        self.subtitle_str = cr.get(u'roadmeet', u'subtitle')
        self.document_str = cr.get(u'roadmeet', u'document')
        self.date_str = cr.get(u'roadmeet', u'date')
        self.organiser_str = cr.get(u'roadmeet', u'organiser')
        self.commissaire_str = cr.get(u'roadmeet', u'commissaire')
        self.distance = cr.get_float(u'roadmeet', u'distance')
        self.diststr = cr.get(u'roadmeet', u'diststr')
        self.docindex = cr.get_posint(u'roadmeet', u'docindex', 0)
        self.linkbase = cr.get(u'roadmeet', u'linkbase')
        self.indexlink = cr.get(u'roadmeet', u'indexlink')
        self.prevlink = cr.get(u'roadmeet', u'prevlink')
        self.nextlink = cr.get(u'roadmeet', u'nextlink')
        self.bibs_in_results = cr.get_bool(u'roadmeet', u'resultnos')
        self.mirrorpath = cr.get(u'roadmeet', u'mirrorpath')
        self.mirrorcmd = cr.get(u'roadmeet', u'mirrorcmd')
        self.mirrorfile = cr.get(u'roadmeet', u'mirrorfile')
        self.competitioncode = cr.get(u'roadmeet', u'competitioncode')
        self.eventcode = cr.get(u'roadmeet', u'eventcode')
        self.racetype = cr.get(u'roadmeet', u'racetype')
        self.competitortype = cr.get(u'roadmeet', u'competitortype')
        self.documentversion = cr.get(u'roadmeet', u'documentversion')
        self.lifexport = cr.get_bool(u'roadmeet', u'lifexport')
        self.resfiles = cr.get_bool(u'roadmeet', u'resfiles')
        self.provisionalstart = cr.get_bool(u'roadmeet', u'provisionalstart')

        # Re-Initialise rider and event databases
        self.rdb.clear()
        self.edb.clear()
        self.rdb.load(u'riders.csv')
        self.edb.load(u'events.csv')
        event = self.edb.getfirst()
        if event is None:  # add a new event of the right type
            event = self.edb.add_empty(evno=u'0')
            event[u'type'] = self.etype
        else:
            self.etype = event[u'type']
            _log.debug(u'Existing event in db: %r', self.etype)
        self.open_event(event)  # always open on load if posible
        self.set_title()

        # alt timer config post event load
        if self.etype == u'irtt':
            self.alttimer.write(u'DTS05.00')
            self.alttimer.write(u'DTF00.01')
        else:
            # assume 1 second gaps at finish
            self.alttimer.write(u'DTF01.00')

        # make sure export path exists
        if not os.path.exists(self.exportpath):
            os.mkdir(self.exportpath)
            _log.info(u'Created export path: %r', self.exportpath)

        # check and warn of config mismatch
        cid = cr.get(u'roadmeet', u'id')
        if cid != ROADMEET_ID:
            _log.warning(u'Meet config mismatch: %r != %r', cid, ROADMEET_ID)

    def get_distance(self):
        """Return race distance in km."""
        return self.distance

    ## Announcer methods (replaces old irc/unt telegraph)
    def cmd_announce(self, command, msg):
        """Announce the supplied message to the command topic."""
        if self.anntopic:
            topic = u'/'.join((self.anntopic, command))
            self.announce.publish(msg, topic)

    def rider_announce(self, rvec):
        """Issue a serialised rider vector to announcer."""
        # Deprecated UNT-style list
        self.cmd_announce(u'rider', u'\x1f'.join(rvec))

    def timer_announce(self, evt, timer=None, source=u''):
        """Send message into announce for remote control."""
        if not self.remote_enable and self.timertopic is not None:
            if timer is None:
                timer = self.timer
            prec = 4
            if timer is self.timer:
                prec = 3  # transponders have reduced precision
            elif u'M' in evt.chan:
                prec = 3
            if evt.source is not None:
                source = evt.source
            tvec = (evt.index, source, evt.chan, evt.refid, evt.rawtime(prec),
                    u'')
            self.announce.publish(u';'.join(tvec), self.timertopic)
        self.rfustat.buttonchg(uiutil.bg_armint)
        self.rfuact = True
        return False

    def mirror_start(self):
        """Create a new mirror thread unless already in progress."""
        if self.mirrorpath and self.mirror is None:
            self.mirror = export.mirror(localpath=os.path.join(u'export', u''),
                                        remotepath=self.mirrorpath,
                                        mirrorcmd=self.mirrorcmd)
            self.mirror.start()
        return False  # for idle_add

    def remote_reset(self):
        """Reset remote input of timer messages."""
        _log.debug(u'Remote control reset')
        if self.timertopic is not None:
            if self.remote_enable:
                _log.debug(u'Listening for remote timer at %r',
                           self.timertopic)
                self.announce.subscribe(self.timertopic)
            else:
                _log.debug(u'Remote timer disabled')
                self.announce.unsubscribe(self.timertopic)
        else:
            _log.debug(u'Remote timer topic not cofigured')

    def remote_timer(self, msg):
        """Process and dispatch a remote timer message."""
        # 'INDEX;SOURCE;CHANNEL;REFID;TIMEOFDAY'
        tv = msg.split(u';')
        if len(tv) == 5 or len(tv) == 6:
            try:
                if len(tv) > 5:
                    # check date against today
                    # if today != tv[5]:
                    # log and return
                    pass
                tval = tod.mktod(tv[4])
                tval.source = tv[1]
                tval.chan = tv[2]
                tval.refid = tv[3]
                _log.debug(u'Remote src:%r index:%r chan:%r refid:%r @ %r',
                           tv[1], tv[0], tv[2], tv[3], tval.rawtime())
                if u'timy' in tv[1]:
                    tval.index = tv[0]
                    self._alttimercb(tval)
                else:
                    tval.index = u'REM'
                    self._timercb(tval)
            except Exception as e:
                _log.warning(u'Error reading timer msg %r: %s', msg, e)
        else:
            _log.debug(u'Invalid remote timer message: %r', tv)

    def remote_command(self, topic, msg):
        """Handle a remote control message."""
        if topic == self.timertopic:
            if self.remote_enable:
                self.remote_timer(msg)
        else:
            _log.debug(u'Unsupported remote command %r:%r', topic, msg)
        return False

    def _timercb(self, evt, data=None):
        if self.timercb is not None:
            glib.idle_add(self.timercb, evt, priority=glib.PRIORITY_HIGH)
        glib.idle_add(self.timer_announce, evt, self.timer, u'rfid')

    def _alttimercb(self, evt, data=None):
        if self.alttimercb is not None:
            glib.idle_add(self.alttimercb, evt, priority=glib.PRIORITY_HIGH)
        glib.idle_add(self.timer_announce, evt, self.alttimer, u'timy')

    def _controlcb(self, topic=None, message=None):
        glib.idle_add(self.remote_command, topic, message)

    def __init__(self, etype=None, lockfile=None):
        """Meet constructor."""
        self.loghandler = None  # set in loadconfig to meet dir
        self.exportpath = EXPORTPATH
        if etype not in ROADRACE_TYPES:
            etype = u'road'
        self.etype = etype
        self.meetlock = lockfile
        self.shortname = None
        self.title_str = u''
        self.host_str = u''
        self.subtitle_str = u''
        self.document_str = u''
        self.date_str = u''
        self.organiser_str = u''
        self.commissaire_str = u''
        self.distance = None
        self.diststr = u''
        self.docindex = 0
        self.linkbase = u'.'
        self.provisionalstart = False
        self.indexlink = None
        self.nextlink = None
        self.prevlink = None

        self.bibs_in_results = True
        self.remote_enable = False
        self.lifexport = False
        self.resfiles = True

        # printer preferences
        paper = gtk.paper_size_new_custom('metarace-full', 'A4 for reports',
                                          595, 842, gtk.UNIT_POINTS)
        self.printprefs = gtk.PrintSettings()
        self.pageset = gtk.PageSetup()
        self.pageset.set_orientation(gtk.PAGE_ORIENTATION_PORTRAIT)
        self.pageset.set_paper_size(paper)
        self.pageset.set_top_margin(0, gtk.UNIT_POINTS)
        self.pageset.set_bottom_margin(0, gtk.UNIT_POINTS)
        self.pageset.set_left_margin(0, gtk.UNIT_POINTS)
        self.pageset.set_right_margin(0, gtk.UNIT_POINTS)

        # hardware connections
        self.timertopic = None  # remote timer topic
        self.timer = decoder.decoder()
        self.timer_port = u''
        self.timer.setcb(self._timercb)
        self.timercb = None  # set by event app
        self.alttimer = timy.timy()  # alttimer is always timy
        self.alttimer_port = u''
        self.alttimer.setcb(self._alttimercb)
        self.alttimercb = None  # set by event app
        self.announce = telegraph.telegraph()
        self.announce.setcb(self._controlcb)
        self.anntopic = None
        self.mirrorpath = u''
        self.mirrorcmd = u'echo'
        self.mirrorfile = u''
        self.mirror = None
        self.competitioncode = u''
        self.eventcode = u''
        self.racetype = u''
        self.competitortype = u''
        self.documentversion = u''

        uifile = os.path.join(metarace.UI_PATH, u'roadmeet.ui')
        _log.debug(u'Building user interface from %r', uifile)
        b = gtk.Builder()
        b.add_from_file(uifile)
        self.window = b.get_object(u'meet')
        self.window.connect(u'key-press-event', self.key_event)
        self.rfustat = uiutil.statbut(b.get_object(u'menu_clock'))
        self.rfustat.buttonchg(uiutil.bg_none, '--')
        self.rfuact = False
        self.rfufail = 0
        self.status = b.get_object(u'status')
        self.log_buffer = b.get_object(u'log_buffer')
        self.log_view = b.get_object(u'log_view')
        self.log_view.modify_font(uiutil.LOGVIEWFONT)
        self.log_scroll = b.get_object(u'log_box').get_vadjustment()
        self.context = self.status.get_context_id(u'metarace meet')
        self.menu_race_close = b.get_object(u'menu_race_close')
        self.menu_race_abort = b.get_object(u'menu_race_abort')
        self.decoder_configure = b.get_object(u'menu_timing_configure')
        self.race_box = b.get_object(u'race_box')
        self.stat_but = uiutil.statbut(b.get_object(u'race_stat_but'))
        self.action_model = b.get_object(u'race_action_model')
        self.action_combo = b.get_object(u'race_action_combo')
        self.action_entry = b.get_object(u'race_action_entry')
        b.get_object(u'race_stat_hbox').set_focus_chain(
            [self.action_combo, self.action_entry, self.action_combo])

        # prepare local scratch pad ? can these be removed?
        self.an_cur_lap = tod.ZERO
        self.an_cur_split = tod.ZERO
        self.an_cur_bunchid = 0
        self.an_cur_bunchcnt = 0
        self.an_last_time = None
        self.an_cur_start = tod.ZERO

        b.connect_signals(self)

        # run state
        self.running = True
        self.started = False
        self.curevent = None

        # connect UI log handlers
        _log.debug(u'Connecting interface log handlers')
        rootlogger = logging.getLogger()
        f = logging.Formatter(metarace.LOGFORMAT)
        self.sh = loghandler.statusHandler(self.status, self.context)
        self.sh.setFormatter(f)
        self.sh.setLevel(logging.INFO)  # show info+ on status bar
        rootlogger.addHandler(self.sh)
        self.lh = loghandler.textViewHandler(self.log_buffer, self.log_view,
                                             self.log_scroll)
        self.lh.setFormatter(f)
        self.lh.setLevel(logging.INFO)  # show info+ in text view
        rootlogger.addHandler(self.lh)

        # get rider db and pack into a dialog
        _log.debug(u'Add riderdb and eventdb')
        self.rdb = riderdb.riderdb()
        b.get_object(u'riders_box').add(
            self.rdb.mkview(cat=True,
                            series=True,
                            refid=True,
                            ucicode=True,
                            note=True))
        # get event db -> loadconfig adds empty event if one not present
        self.edb = eventdb.eventdb([])

        # select event page in notebook.
        b.get_object(u'meet_nb').set_current_page(0)

        # start timer
        glib.timeout_add_seconds(1, self.timeout)


def main():
    """Run the road meet application as a console script."""
    # attach a console log handler to the root logger then run the ui
    ch = logging.StreamHandler()
    ch.setLevel(metarace.LOGLEVEL)
    fh = logging.Formatter(metarace.LOGFORMAT)
    ch.setFormatter(fh)
    logging.getLogger().addHandler(ch)

    metarace.init(withgtk=True)
    configpath = metarace.DATA_PATH
    if len(sys.argv) > 2:
        _log.error(u'Usage: roadmeet [configdir]')
        sys.exit(1)
    elif len(sys.argv) == 2:
        configpath = sys.argv[1]
    configpath = metarace.config_path(configpath)
    if configpath is None:
        _log.error(u'Unable to open meet config %r', sys.argv[1])
        sys.exit(1)
    app = runapp(configpath)
    try:
        metarace.mainloop()
    except:
        app.shutdown(u'Exception from main loop')
        raise
    return 0


def runapp(configpath, etype=None):
    """Create the roadmeet object, start in configpath and return a handle."""
    lf = metarace.lockpath(configpath)
    if lf is None:
        _log.error(u'Unable to lock meet config, already in use')
        sys.exit(1)
    _log.debug(u'Entering meet folder %r', configpath)
    os.chdir(configpath)
    app = roadmeet(etype, lf)
    app.loadconfig()
    app.window.show()
    app.start()
    return app


if __name__ == u'__main__':
    main()



"""Rider 'database' and utilities

Manage a legacy rider db with the columns described below.
Riders are uniquely identified by their series and number.
Columns must all appear in the order listed.

Legacy columns:

	No	Rider number / team code
	First	First name / team name
	Last	Last name / team short name
	Org	Club name or Team Code
	Cat	Result categories
	Series	Number series
	Refid	Transponder ID
	UCIID	UCI ID / Team dataride ID
	DoB	Date of Birth
	Nation	Sporting Nationality
	Sex	Rider sex / team category
	Note	Spare data slot

"""

import gtk
import gobject
import glib
import logging
import os
from unicodedata import normalize

import metarace
from metarace import uiutil
from metarace import strops
from metarace import ucsv
from metarace import namebank

_log = logging.getLogger(u'metarace.riderdb')
_log.setLevel(logging.DEBUG)

# Model column constants (unkeyed csv)
COL_BIB = 0
COL_FIRST = 1
COL_LAST = 2
COL_ORG = 3
COL_CAT = 4
COL_SERIES = 5
COL_REFID = 6
COL_UCICODE = 7
COL_DOB = 8
COL_NATION = 9
COL_SEX = 10
COL_NOTE = 11

# Reserved rider numbers.
RESERVED_NOS = [u'bib', u'no.', u'no', u'all', u'number', u'num']

# Default rider values if not empty string
RIDER_DEFAULTS = {
    u'no': None,
}


def primary_cat(catstr):
    """Return the primary category from a rider cat list."""
    ret = u''
    cv = catstr.split()
    if cv:
        ret = cv[0].upper()
    return ret


def colkey(colstr=u''):
    return colstr[0:4].lower().strip()


def cellnorm(unistr):
    """Normalise supplied string, then return a version with printing chars."""
    return normalize('NFC', unistr).translate(strops.PRINT_UTRANS)


class riderdb(object):
    """Rider database."""

    def addempty(self, bib=u'', series=u''):
        """Add a new empty row in the rider model."""
        i = self.model.append([
            bib.lower(), u'', u'', u'', u'', series, u'', u'', u'', u'', u'',
            u''
        ])
        ref = gtk.TreeRowReference(self.model, self.model.get_path(i))
        if self.view is not None:
            self.postedit = None
            self.gotorow(ref)
        return ref

    def clear(self):
        """Clear rider model."""
        _log.debug('Rider model cleared')
        self.model.clear()

    def load_chipfile(self, csvfile=None):
        """Load refids from the named chipfile into to the riderdb."""
        count = 0
        if os.path.isfile(csvfile):
            _log.debug(u'Loading refids from %r', csvfile)
            with open(csvfile, 'rb') as f:
                cr = ucsv.UnicodeReader(f)
                for row in cr:
                    ir = [cellnorm(cell) for cell in row]
                    if len(ir) > 1 and ir[0] != u'Refid':
                        nrefid = ir[0].strip().lower()
                        bib = ir[1].strip().lower()
                        series = u''
                        if len(ir) > 2:
                            series = ir[2].strip()

                        if bib.isalnum() and bib not in RESERVED_NOS:
                            bibser = strops.bibser2bibstr(bib, series)
                            self.tagmap[nrefid] = bibser
                            self.maptag[bibser] = nrefid
                            dbr = self.getrider(bib, series)
                            if dbr is not None:
                                self.editrider(ref=dbr, refid=nrefid)
                            count += 1
                        else:
                            _log.warning(u'Invalid rider %r', bib)
            _log.debug(u'Loaded %r refids from %r', count, csvfile)

    def load(self, csvfile=None, namedb=None, overwrite=False):
        """Load riders from supplied CSV file."""
        count = 0
        if os.path.isfile(csvfile):
            _log.debug(u'Loading riders from %r', csvfile)
            with open(csvfile, 'rb') as f:
                cr = ucsv.UnicodeReader(f)
                for row in cr:
                    ir = [cellnorm(cell) for cell in row]
                    if len(ir) > 0 and ir[0] != u'Bib' and ir[0] != u'No.':
                        bib = ir[COL_BIB].strip().lower()
                        if bib.isalnum() and bib not in RESERVED_NOS:
                            nr = [
                                bib, u'', u'', u'', u'', u'', u'', u'', u'',
                                u'', u'', u''
                            ]
                            for i in range(1, 12):
                                if len(ir) > i:
                                    nr[i] = ir[i].strip()

                            # overwrite refid in load data if chipfile loaded
                            bibser = strops.bibser2bibstr(bib, nr[COL_SERIES])
                            if bibser in self.maptag:
                                if nr[COL_REFID] == u'':
                                    nr[COL_REFID] = self.maptag[bibser]
                            oldr = self.getrider(bib, nr[COL_SERIES])
                            if oldr is not None:
                                # entry already exists, ignore or remove
                                if overwrite:
                                    _log.info(u'Replaced %r', bibser)
                                    self.model.remove(
                                        self.model.get_iter(oldr.get_path()))
                                else:
                                    _log.warning(u'Duplicate ignored %r',
                                                 bibser)
                                    continue
                            self.model.append(nr)
                            count += 1
                        else:
                            _log.warning(u'Invalid rider %r', bib)
            _log.debug('Loaded %r riders from %r', count, csvfile)

    def save_chipfile(self, csvfile=None):
        """Save refids from current model to supplied CSV file."""
        chipmap = {}
        chiplist = []
        count = 0
        if os.path.exists(csvfile):
            # load in old chipfile
            with open(csvfile, 'rb') as f:
                cr = ucsv.UnicodeReader(f)
                for row in cr:
                    ir = [cell.translate(strops.PRINT_UTRANS) for cell in row]
                    if len(ir) > 1 and ir[0] != u'Refid':
                        nrefid = ir[0].strip().lower()
                        bib = ir[1].strip().lower()
                        series = u''
                        if len(ir) > 2:
                            series = ir[2].strip()

                        if bib.isalnum() and bib not in RESERVED_NOS:
                            bibstr = strops.bibser2bibstr(bib, series)
                            if bibstr not in chipmap:
                                chipmap[bibstr] = nrefid
                                chiplist.append(bibstr)
                                count += 1
                        else:
                            _log.warning(u'Invalid rider %r', bib)
                _log.debug(u'Loaded %r refids from %r', count, csvfile)

        # overwrite any existing refids
        ocount = 0
        ncount = 0
        for r in self.model:
            urefid = r[COL_REFID].decode('utf-8')
            if urefid:
                ubib = r[COL_BIB].decode('utf-8')
                useries = r[COL_SERIES].decode('utf-8')
                bibstr = strops.bibser2bibstr(ubib, useries)
                if bibstr not in chipmap:
                    chiplist.append(bibstr)
                    ncount += 1
                else:
                    if urefid != chipmap[bibstr]:
                        ocount += 1
                chipmap[bibstr] = urefid
        _log.debug(u'Replacing %r refids in model', ocount)
        _log.debug(u'Adding %r new refids to model', ncount)

        chiplist.sort(key=strops.bibstr_key)
        count = 0
        with open(csvfile, 'wb') as f:
            cr = ucsv.UnicodeWriter(f)
            cr.writerow([u'Refid', u'No.', u'Series'])
            for bs in chiplist:
                (bib, ser) = strops.bibstr2bibser(bs)
                refid = chipmap[bs]
                if refid:
                    count += 1
                    cr.writerow([refid, bib, ser])
        _log.debug(u'Wrote %r refids to %r', count, csvfile)

    def save(self, csvfile=None):
        """Save current model to supplied CSV file."""
        _log.debug(u'Saving riders to %r', csvfile)
        with metarace.savefile(csvfile) as f:
            cr = ucsv.UnicodeWriter(f)
            cr.writerow([
                u'No.', u'First Name', u'Last Name', u'Org', u'Category',
                u'Series', u'Refid', u'UCI ID', u'DoB', u'Nationality', u'Sex',
                u'Note'
            ])
            cr.writerows(self)  # check the unicode out-in-out-in

    def gotorow(self, ref=None):
        """Move view selection to the specified row reference."""
        if ref is None and len(self.model) > 0:
            ref = gtk.TreeRowReference(self.model, 0)
        if ref is not None and ref.valid():
            path = ref.get_path()
            self.view.scroll_to_cell(path)
            self.view.set_cursor_on_cell(path)

    def delselected(self):
        """Delete the currently selected row."""
        if self.view is not None:
            model, iter = self.view.get_selection().get_selected()
            if iter is not None:
                ref = None
                if self.model.remove(iter):
                    ref = gtk.TreeRowReference(self.model,
                                               self.model.get_path(iter))
                self.gotorow(ref)

    def getselected(self):
        """Return a reference to the currently selected row, or None."""
        ref = None
        if self.view is not None:
            model, iter = self.view.get_selection().get_selected()
            if iter is not None:
                ref = gtk.TreeRowReference(self.model,
                                           self.model.get_path(iter))
        return ref

    def mkridermap(self, series=u''):
        """Return a formatted rider map for mkresult scripts."""
        # Note: Hack until riderdb modified to be same as eventdb
        ret = {}
        for r in self.model:
            if r[COL_SERIES] == series:
                key = r[COL_BIB].decode('utf-8')
                first = r[COL_FIRST].decode('utf-8')
                last = r[COL_LAST].decode('utf-8')
                org = r[COL_ORG].decode('utf-8')
                cat = r[COL_CAT].decode('utf-8')
                refid = r[COL_REFID].decode('utf-8')
                uciid = r[COL_UCICODE].decode('utf-8')
                dob = r[COL_DOB].decode('utf-8')
                nation = r[COL_NATION].decode('utf-8')
                sex = r[COL_SEX].decode('utf-8')
                note = r[COL_NOTE].decode('utf-8')
                rmap = {}
                rmap[u'bib'] = key
                rmap[u'first'] = first
                rmap[u'last'] = last
                rmap[u'org'] = org
                rmap[u'cat'] = cat
                rmap[u'refid'] = refid
                rmap[u'uciid'] = uciid
                rmap[u'dob'] = dob
                rmap[u'nation'] = nation
                rmap[u'sex'] = sex
                rmap[u'note'] = note
                rmap[u'namestr'] = strops.resname(first, last, org)
                rmap[u'name'] = strops.resname(first, last)
                rmap[u'shortname'] = strops.fitname(first,
                                                    last,
                                                    4,
                                                    trunc=False)
                ret[key] = rmap
        return ret

    def mkview(self,
               bib=True,
               first=True,
               last=True,
               club=True,
               cat=False,
               series=True,
               refid=False,
               ucicode=False,
               note=False):
        """Create and return view object for the model."""
        if self.view is not None:
            return self.view
        v = gtk.TreeView(self.model)
        v.set_reorderable(True)
        v.set_enable_search(False)
        v.set_rules_hint(True)
        v.connect('key-press-event', self.__view_key)
        v.show()
        self.colmap = {}
        colcnt = 0
        if bib:
            uiutil.mkviewcoltxt(v,
                                'No.',
                                COL_BIB,
                                self.__editcol_cb,
                                halign=0.5,
                                calign=0.5,
                                editcb=self.__editstart_cb)
            self.colmap[COL_BIB] = colcnt
            colcnt += 1
        if first:
            uiutil.mkviewcoltxt(v,
                                'First Name',
                                COL_FIRST,
                                self.__editcol_cb,
                                expand=True,
                                editcb=self.__editstart_cb)
            self.colmap[COL_FIRST] = colcnt
            colcnt += 1
        if last:
            uiutil.mkviewcoltxt(v,
                                'Last Name',
                                COL_LAST,
                                self.__editcol_cb,
                                expand=True,
                                editcb=self.__editstart_cb)
            self.colmap[COL_LAST] = colcnt
            colcnt += 1
        if club:
            uiutil.mkviewcoltxt(v,
                                'Org',
                                COL_ORG,
                                self.__editcol_cb,
                                editcb=self.__editstart_cb)
            self.colmap[COL_ORG] = colcnt
            colcnt += 1
        if cat:
            uiutil.mkviewcoltxt(v,
                                'Cat',
                                COL_CAT,
                                self.__editcol_cb,
                                editcb=self.__editstart_cb)
            self.colmap[COL_CAT] = colcnt
            colcnt += 1
        if series:
            uiutil.mkviewcoltxt(v,
                                'Ser',
                                COL_SERIES,
                                self.__editcol_cb,
                                editcb=self.__editstart_cb)
            self.colmap[COL_SERIES] = colcnt
            colcnt += 1
        if refid:
            uiutil.mkviewcoltxt(v,
                                'Refid',
                                COL_REFID,
                                self.__editcol_cb,
                                editcb=self.__editstart_cb)
            self.colmap[COL_REFID] = colcnt
            colcnt += 1
        if ucicode:
            uiutil.mkviewcoltxt(v,
                                'UCI ID',
                                COL_UCICODE,
                                self.__editcol_cb,
                                editcb=self.__editstart_cb)
            self.colmap[COL_UCICODE] = colcnt
            colcnt += 1
        if note:
            uiutil.mkviewcoltxt(v,
                                'Note',
                                COL_NOTE,
                                self.__editcol_cb,
                                editcb=self.__editstart_cb)
            self.colmap[COL_NOTE] = colcnt
            colcnt += 1
        self.view = v
        return self.view

    def getrider(self, bib, series=u''):
        """Return a reference to the row with the given bib and series."""
        ret = None
        bib = bib.lower()
        i = self.model.get_iter_first()
        while i is not None:
            if (self.get_value(i, COL_BIB) == bib
                    and self.get_value(i, COL_SERIES) == series):
                ret = gtk.TreeRowReference(self.model, self.model.get_path(i))
                break
            i = self.model.iter_next(i)
        return ret

    def get_partner(self, bib, series=u''):
        """Return a reference to the 'partner' for the specified rider."""
        ret = None
        bib = bib.lower()
        i = self.model.get_iter_first()
        while i is not None:
            if (self.get_value(i, COL_BIB) == bib
                    and self.get_value(i, COL_SERIES) == series):
                ret = gtk.TreeRowReference(self.model, self.model.get_path(i))
                break
            i = self.model.iter_next(i)

        if i is not None:
            # look up cat and note for partner link
            cat = self.get_value(i, COL_CAT)
            if cat and u'tandem' in cat.lower():
                pno = self.get_value(i, COL_NOTE)
                if pno:
                    ret = self.getrider(pno, series)
        return ret

    def nextriderin(self, bib, series=u''):
        """Return the rider following the specified bib in the series."""
        ret = None
        cur = self.getrider(bib, series)
        if cur is not None:
            i = self.model.iter_next(self.model.get_iter(cur.get_path()))
            while i is not None:
                if self.get_value(i, COL_SERIES) == series:
                    ret = self.get_value(i, COL_BIB)
                    break
                i = self.model.iter_next(i)
        return ret

    def nextriderno(self):
        """Try and return a new unique rider number string."""
        lmax = 1
        for r in self.model:
            ubib = r[COL_BIB].decode('utf-8')
            if ubib.isdigit() and int(ubib) >= lmax:
                lmax = int(ubib) + 1
        return unicode(lmax)

    def getrefid(self, refid):
        """Return a reference to the row with the given refid."""
        ## altered to allow multiple references
        ret = None
        ck = refid.lower()
        rno = None
        rser = None
        if u'riderno:' in ck:
            rno, rser = strops.bibstr2bibser(ck.split(u':')[-1])
        i = self.model.get_iter_first()
        while i is not None:
            if rno and (self.get_value(i, COL_BIB) == rno
                        and self.get_value(i, COL_SERIES) == rser):
                ret = gtk.TreeRowReference(self.model, self.model.get_path(i))
                break
            else:
                for dbref in self.get_value(i, COL_REFID).lower().split():
                    if dbref == ck:
                        ret = gtk.TreeRowReference(self.model,
                                                   self.model.get_path(i))
                        break
            i = self.model.iter_next(i)
        return ret

    def biblistfromcat(self, cat, series=u''):
        """Return a list of rider numbers in the given cat."""
        ret = []
        cat = cat.upper()
        i = self.model.get_iter_first()
        while i is not None:
            if (cat in self.get_value(i, COL_CAT).upper().split()
                    and self.get_value(i, COL_SERIES) == series):
                ret.append(self.get_value(i, COL_BIB))
            i = self.model.iter_next(i)
        return ret

    def getbibs(self, cat, series=u''):
        """Return a list of refs to riders in the given cat and series."""
        ret = []
        cat = cat.upper()
        i = self.model.get_iter_first()
        while i is not None:
            if (cat in self.get_value(i, COL_CAT).upper().split()
                    and self.get_value(i, COL_SERIES) == series):
                ret.append(
                    gtk.TreeRowReference(self.model, self.model.get_path(i)))
            i = self.model.iter_next(i)
        return ret

    def listseries(self):
        """Return a list of all the series in the rider db."""
        ret = []
        for row in self.model:
            useries = row[COL_SERIES].decode('utf-8')
            if useries not in ret:
                ret.append(useries)
        return ret

    def listcats(self, series=None):
        """Return a list of all the categories in the specified series."""
        ret = []
        for row in self.model:
            ucat = row[COL_CAT].decode('utf-8').upper().split()
            useries = row[COL_SERIES].decode('utf-8')
            if useries not in [u'cat', u'spare', u'team']:
                for rrcat in ucat:
                    if rrcat not in [u'SPARE', u'CAT', u'TEAM']:
                        if rrcat and rrcat not in ret:
                            if series is None or useries == series:
                                ret.append(rrcat)
        return ret

    def get_value(self, iter, column):
        return self.model.get_value(iter, column).decode('utf-8')

    def get_rowvalue(self, row, column):
        return row[column].decode('utf-8')

    def getvalue(self, ref, col):
        """Return the specified column from the supplied row."""
        ret = None
        if ref.valid():
            ret = self.model[ref.get_path()][col].decode('utf-8')
        return ret

    def editrider(self,
                  ref=None,
                  first=None,
                  last=None,
                  org=None,
                  cat=None,
                  refid=None,
                  ucicode=None,
                  dob=None,
                  nation=None,
                  sex=None,
                  note=None,
                  bib=None):
        """Create or update the rider with supplied parameters."""
        i = None
        if ref is None:
            if bib is None:
                bib = self.nextriderno()
            ## ERROR: num not ever defined?
            i = self.model.append(
                [bib, u'', u'', u'', u'', u'', u'', u'', u''])
            ref = gtk.TreeRowReference(self.model, self.model.get_path(i))
        if ref.valid():
            i = self.model.get_iter(ref.get_path())
            if first is not None:
                self.model.set_value(i, COL_FIRST, first)
            if last is not None:
                self.model.set_value(i, COL_LAST, last)
            if org is not None:
                self.model.set_value(i, COL_ORG, org)
            if cat is not None:
                self.model.set_value(i, COL_CAT, cat)
            if refid is not None:
                self.model.set_value(i, COL_REFID, refid)
            if ucicode is not None:
                self.model.set_value(i, COL_UCICODE, ucicode)
            if dob is not None:
                self.model.set_value(i, COL_DIB, dob)
            if nation is not None:
                self.model.set_value(i, COL_NATION, nation)
            if sex is not None:
                self.model.set_value(i, COL_SEX, sex)
            if note is not None:
                self.model.set_value(i, COL_NOTE, note)
        return ref

    def filter_nonempty(self, col=None):
        """Return a filter model with col non empty."""
        ret = self.model.filter_new()
        ret.set_visible_func(self.__filtercol, col)
        return ret

    def __editcol_cb(self, cell, path, new_text, col):
        """Update model if possible and request post-edit movement."""
        ret = False
        if new_text is not None:
            new_text = new_text.decode(u'utf-8').strip()
            if new_text != self.model[path][col].decode(u'utf-8'):
                if col == COL_BIB:
                    if new_text.isalnum() and new_text.lower != u'all':
                        if not self.getrider(
                                new_text,
                                self.model[path][COL_SERIES].decode(u'utf-8')):
                            self.model[path][COL_BIB] = new_text.lower()
                            ret = True
                        else:
                            _log.warning(u'Rider %r already exists', new_text)
                            self.postedit = u'same'
                            ret = True  # re-focus on the entry
                    else:
                        _log.warning(u'Invalid number %r ignored', new_text)
                elif col == COL_SERIES:
                    if not self.getrider(
                            self.model[path][COL_BIB].decode(u'utf-8'),
                            new_text):
                        self.model[path][COL_SERIES] = new_text
                        ret = True
                    else:  # This path is almost never a real problem
                        _log.debug(u'Did not update series to duplicate rider')
                        ret = True
                elif col == COL_CAT:
                    nt = u' '.join(new_text.replace(',', ' ').split())
                    self.model[path][col] = nt
                    ret = True
                else:
                    self.model[path][col] = new_text
                    ret = True
            else:  # No Change, but entry 'accepted'
                ret = True
        if ret and self.postedit is not None:
            glib.idle_add(self.__postedit_move, path, self.colmap[col])
        return ret

    def __postedit_move(self, path, col):
        """Perform a post-edit movement of the selection.

           NOTE: This 'col' refers to VIEW column index and not the
                 model's column index.

        """
        if self.postedit is None:  # race possible here
            return False

        # step 1: process left/right
        if self.postedit == 'left':
            col -= 1
            self.postedit = None
        elif self.postedit == 'right':
            col += 1
            self.postedit = None

        if col < 0:
            col = len(self.colmap) - 1
            self.postedit = 'up'  # followup with a upward movement
        elif col >= len(self.colmap):
            col = 0
            self.postedit = 'down'  # followup with a downward movement

        # step 2: check for additional up/down
        i = self.model.get_iter(path)
        if self.postedit == 'up':
            p = int(self.model.get_string_from_iter(i)) - 1
            if p >= 0:
                path = self.model.get_path(
                    self.model.get_iter_from_string(str(p)))
            else:
                return False  # can't move any further 'up'
        elif self.postedit == 'down':
            i = self.model.iter_next(i)
            if i is not None:
                path = self.model.get_path(i)
            else:
                return False  # no more rows to scroll to -> perhaps add?
        self.postedit = None  # suppress any further change
        return self.__moveto_col(path, col)

    def __moveto_col(self, path, col):
        """Move selection to supplied path and column.

           NOTE: This 'col' refers to VIEW column index and not the
                 model's column index.

        """
        self.view.scroll_to_cell(path, self.view.get_column(col), False)
        self.view.set_cursor(path, self.view.get_column(col), True)
        return False

    def __view_key(self, widget, event):
        """Handle key events on tree view."""
        if event.type == gtk.gdk.KEY_PRESS:
            if event.state & gtk.gdk.CONTROL_MASK:
                key = gtk.gdk.keyval_name(event.keyval) or 'None'
                if key.lower() == 'a':
                    self.addempty()
                    return True
                elif key == 'Delete':
                    self.delselected()
                    return True
        return False

    def __edit_entry_key_cb(self, widget, event, editable=None):
        """Check key press in cell edit for postedit move."""
        if event.type == gtk.gdk.KEY_PRESS:
            key = gtk.gdk.keyval_name(event.keyval) or 'None'
            if key == 'Tab':
                self.postedit = 'right'
            elif key in ['Return', 'Escape']:  # allow cancel to handle
                self.postedit = None
            elif key == 'Up':
                if editable is None:
                    self.postedit = 'up'
            elif key == 'Down':
                if editable is None:
                    self.postedit = 'down'
            elif key == 'Right':
                if self.editwasempty:
                    self.postedit = 'right'
                    if editable is not None:
                        editable.editing_done()
                    else:
                        widget.editing_done()
            elif key == 'Left':
                if self.editwasempty:
                    self.postedit = 'left'
                    if editable is not None:
                        editable.editing_done()
                    else:
                        widget.editing_done()
        return False

    def __editstart_cb(self, cr, editable, path, data=None):
        """Prepare cell entry for post-edit movement."""
        self.postedit = None
        if type(editable) is gtk.Entry:
            self.editwasempty = len(editable.get_text()) == 0
            editable.connect('key-press-event', self.__edit_entry_key_cb)
        else:
            self.editwasempty = False

    def __filtercol(self, model, iter, data=None):
        return bool(self.model.get_value(iter, data))

    def __iter__(self):
        """Map out unicode strings."""
        for r in self.model:
            yield [a.decode('utf-8') for a in r]

    def __init__(self):
        """Constructor for the rider db.

        Constructs a new rider database object. This function does
        not create the view object, use the mkview() function to
        create and return a valid treeview.

        """

        self.model = gtk.ListStore(
            gobject.TYPE_STRING,  # 0 bib
            gobject.TYPE_STRING,  # 1 first name
            gobject.TYPE_STRING,  # 2 last name
            gobject.TYPE_STRING,  # 3 org
            gobject.TYPE_STRING,  # 4 category
            gobject.TYPE_STRING,  # 5 series
            gobject.TYPE_STRING,  # 6 refid
            gobject.TYPE_STRING,  # 7 ucicode
            gobject.TYPE_STRING,  # 8 dob
            gobject.TYPE_STRING,  # 9 nation
            gobject.TYPE_STRING,  # 10 sex
            gobject.TYPE_STRING)  # 11 note
        self.tagmap = {}
        self.maptag = {}
        self.view = None
        self.colvec = []
        self.postedit = None
        self.editwasempty = False


class listofentry(object):
    """Load and manage a MR style list of entries."""

    def __init__(self):
        __model = {}
        __event_name = None



# SPDX-License-Identifier: MIT
"""JSON Configuration module.

  Provides a thin wrapper on a dictionary-based configuration
  with JSON export and import. The structure for a configuration
  is a dictionary of sections, each of which contains a dictionary
  of key/value pairs, where the key is a unicode string and the
  value may be any base type supported by python & JSON.
"""

import json


class config(object):

    def __init__(self, default={}):
        """Create config object with a deep copy of the provided default."""
        self.__store = {}
        for section in default:
            self.__store[section] = {}
            for key in default[section]:
                self.__store[section][key] = default[section][key]

    def __str__(self):
        return json.dumps(self.__store)

    def __unicode__(self):
        return unicode(self.__str__())

    def __repr__(self):
        return u'config({})'.format(repr(self.__store))

    def add_section(self, section):
        if isinstance(section, basestring):
            if section not in self.__store:
                self.__store[section] = dict()
        else:
            raise TypeError(u'Invalid section key: ' + repr(section))

    def has_section(self, section):
        return section in self.__store

    def has_option(self, section, key):
        return section in self.__store and key in self.__store[section]

    def sections(self):
        for sec in self.__store:
            yield sec

    def options(self, section):
        for opt in self.__store[section]:
            yield opt

    def get(self, section, key):
        return self.__store[section][key]

    def set(self, section, key, value):
        if isinstance(key, basestring):
            self.__store[section][key] = value
        else:
            raise TypeError(u'Invalid option key: ' + repr(key))

    def get_float(self, section, key, default=None):
        ret = default
        try:
            ret = float(self.get(section, key))
        except Exception:
            pass
        return ret

    def get_posint(self, section, key, default=None):
        ret = default
        try:
            ret = int(self.get(section, key))
            if ret < 0:
                ret = default
        except Exception:
            pass
        return ret

    def get_int(self, section, key, default=None):
        ret = default
        try:
            ret = int(self.get(section, key))
        except Exception:
            pass
        return ret

    def get_str(self, section, key, default=None):
        ret = self.get(section, key)
        if not isinstance(ret, basestring):
            ret = default
        return ret

    def get_bool(self, section, key):
        temp = self.get(section, key)
        if isinstance(temp, basestring):
            if temp.lower() in [u'yes', u'true', u'1']:
                return True
            else:
                return False
        else:
            return bool(temp)

    def write(self, file):
        json.dump(self.__store, file, indent=1)

    def dumps(self):
        return json.dumps(self.__store, indent=1)

    def dictcopy(self):
        """Return a copy of the configuration as a dictionary object."""
        return dict(self.__store)

    def merge(self, otherconfig, section=None, key=None):
        """Merge values from otherconfig into self."""
        if not isinstance(otherconfig, config):
            raise TypeError(u'Merge expects jsonconfig object.')
        if key is not None and section is not None:  # single value import
            if otherconfig.has_option(section, key):
                self.set(section, key, otherconfig.get(section, key))
        elif section is not None:
            self.add_section(section)  # force even if not already loaded
            if otherconfig.has_section(section):
                for opt in otherconfig.options(section):
                    self.set(section, opt, otherconfig.get(section, opt))
        else:
            for sec in otherconfig.sections():
                if self.has_section(sec):
                    # in this case, only add sections already defined
                    for opt in otherconfig.options(sec):
                        self.set(sec, opt, otherconfig.get(sec, opt))

    def read(self, file):
        addconf = json.load(file)
        if not isinstance(addconf, dict):
            raise TypeError(u'Configuration file is not dict: ' +
                            addconf.__class__.__name__)
        for sec in addconf:
            thesec = addconf[sec]
            if not isinstance(thesec, dict):
                raise TypeError(u'Configuration section is not dict: ' +
                                thesec.__type__.__name__)
            self.add_section(sec)
            for k in thesec:
                self.set(sec, k, thesec[k])



"""DO NOT USE"""

import os
import logging
import gtk
import glib
import gobject

import metarace
from metarace import scbwin
from metarace import jsonconfig
from metarace import tod
from metarace import uiutil
from metarace import eventdb
from metarace import riderdb
from metarace import strops
from metarace import report

_log = logging.getLogger(u'metarace.aggregate')
_log.setLevel(logging.DEBUG)

# config version string
EVENT_ID = u'agg-2.0'

# Model columns (for the view only)
COL_BIB = 0
COL_NAME = 1
COL_CLUB = 2
COL_CAT = 3
COL_SEX = 4
COL_POINTS = 5

# scb function key mappings
key_reannounce = u'F4'  # (+CTRL)
key_abort = u'F5'  # (+CTRL)
key_startlist = u'F3'
key_results = u'F4'


class aggregate(object):

    def loadconfig(self):
        """Load race config from disk."""
        cr = jsonconfig.config({
            u'event': {
                u'id': EVENT_ID,
                u'showinfo': True,
                u'comments': [],
                u'pointspec': u'',
                u'events': u'',
                u'double': u'',
                u'qualifying': u''
            }
        })
        cr.add_section('event')
        if os.path.exists(self.configfile):
            try:
                with open(self.configfile, 'rb') as f:
                    cr.read(f)
            except Exception as e:
                _log.error(u'Unable to read config: %s', e)
        else:
            _log.info(u'%r not found, loading defaults', self.configfile)

        self.update_expander_lbl_cb()
        self.info_expand.set_expanded(
            strops.confopt_bool(cr.get(u'event', u'showinfo')))

        self.pointspec = cr.get(u'event', u'pointspec')
        self.events = cr.get(u'event', u'events')
        self.double = cr.get(u'event', u'double')
        self.qualifying = cr.get(u'event', u'qualifying')
        self.recalculate()  # model is cleared and loaded in recalc
        eid = cr.get(u'event', u'id')
        if eid and eid != EVENT_ID:
            _log.info(u'Event config mismatch: %r != %r', eid, EVENT_ID)

    def startlist_report(self, program=False):
        """Return a startlist report."""
        ret = []
        return ret
        # Don't do a startlist for agg entries

    def get_startlist(self):
        """Return a list of bibs in the rider model."""
        ret = []
        for r in self.riders:
            ret.append(r[COL_BIB].decode(u'utf-8'))
        return u' '.join(ret)

    def saveconfig(self):
        """Save race to disk."""
        if self.readonly:
            _log.error(u'Attempt to save readonly event')
            return
        cw = jsonconfig.config()
        cw.add_section(u'event')
        cw.set(u'event', u'pointspec', self.pointspec)
        cw.set(u'event', u'events', self.events)
        cw.set(u'event', u'double', self.double)
        cw.set(u'event', u'qualifying', self.qualifying)
        cw.set(u'event', u'comments', self.comments)
        cw.set(u'event', u'showinfo', self.info_expand.get_expanded())
        cw.set(u'event', u'id', EVENT_ID)
        _log.debug(u'Saving event config %r', self.configfile)
        with metarace.savefile(self.configfile) as f:
            cw.write(f)

    def result_gen(self):
        """Generator function to export a final result."""
        # don't allow result propagation for aggregate (yet?)
        for r in []:
            yield ['', '', None, None]

    def result_report(self, recurse=True):  # by default include inners
        """Return a list of report sections containing the race result."""
        ret = []
        after = u''
        if self.afterev is not None:
            after = ' After Event {}'.format(self.afterev)

        # Club Result (with averages) - mark avg winner with chevron
        sec = report.section(u'teams')
        if self.event[u'evov']:
            sec.heading = u' '.join([self.event[u'pref'],
                                     self.event[u'info']]).strip()
        else:
            sec.heading = u'Event ' + self.evno + u': ' + u' '.join(
                [self.event[u'pref'], self.event[u'info']]).strip()
        sec.subheading = u'Team Tally' + after
        sec.lines = self.teamres
        sec.colheader = [None, None, None, None, u'Avg', u'Total']
        sec.units = u'pt'
        ret.append(sec)

        # Individual Tally - mark sex and cat
        sec = report.section(u'individual')
        sec.subheading = u'Individual Tally' + after
        sec.lines = self.riderres
        sec.units = u'pt'
        ret.append(sec)

        return ret

        # start with the overall result
        sec = report.section()
        if self.event[u'evov']:
            sec.heading = u' '.join([self.event[u'pref'],
                                     self.event[u'info']]).strip()
        else:
            sec.heading = u'Event ' + self.evno + u': ' + u' '.join(
                [self.event[u'pref'], self.event[u'info']]).strip()
        sec.lines = []
        lapstring = strops.lapstring(self.event[u'laps'])
        substr = u' '.join(
            [lapstring, self.event[u'dist'], self.event[u'prog']]).strip()
        if substr:
            sec.subheading = substr
        prevmedal = u''

    def addrider(self, bib=u'', place=u''):
        """Add specified rider to race model."""
        # note this is not relevant for aggregate
        return None

    def getrider(self, bib):
        """Return temporary reference to model row."""
        # note this is not relevant for aggregate
        return None

    def delrider(self, bib):
        """Remove the specified rider from the model."""
        # note this is not relevant for aggregate
        return None

    def getiter(self, bib):
        """Return temporary iterator to model row."""
        # note this is not relevant for aggregate
        return None

    def recalculate(self):
        """Update internal model."""

        # clear out view and data structures
        self.riders.clear()
        self.ridermap = {}
        self.riderres = []
        self.teammap = {}
        self.teamres = []
        self.afterev = None

        # check config (?)
        events = self.events.split()
        doubles = self.double.split()
        points = [int(i) for i in self.pointspec.split()]

        # collect points and allocate to riders and teams
        for evno in events:
            r = self.meet.get_event(evno, False)
            if r is None:
                _log.warning(u'Event %r not found', evno)
                return
            r.loadconfig()  # now have queryable event handle
            double = False
            if evno in doubles:
                double = True
            for res in r.result_gen():
                if isinstance(res[1], int):
                    self.afterev = evno
                    oft = res[1] - 1
                    pts = 0
                    if oft < len(points):
                        pts = points[oft]
                    else:
                        _log.debug(u'More places than points: %r > %r', oft,
                                   points)
                    if double:
                        pts += pts

                    riderno = res[0]
                    if riderno not in self.ridermap:
                        name = u''
                        club = u''
                        namestr = u''
                        cat = u''
                        sex = u''
                        rh = self.meet.newgetrider(riderno)
                        if rh is not None:
                            name = rh[u'name']
                            club = rh[u'club']
                            namestr = rh[u'namestr']
                            cat = rh[u'cat']
                            sex = rh[u'ucicode']
                        self.ridermap[riderno] = {
                            u'name': name,
                            u'club': club,
                            u'namestr': namestr,
                            u'cat': cat,
                            u'sex': sex,
                            u'points': 0
                        }
                    self.ridermap[riderno][u'points'] += pts
                    club = self.ridermap[riderno][u'club']
                    if club not in self.teammap:
                        self.teammap[club] = {
                            u'name': club,
                            u'riders': set(),
                            u'points': 0,
                        }
                    self.teammap[club][u'riders'].add(riderno)
                    self.teammap[club][u'points'] += pts

        # create an aux sorting for overall points
        rideraux = []
        for rider in self.ridermap:
            pts = self.ridermap[rider][u'points']
            riderval = strops.bibstr_key(rider)
            rideraux.append((-pts, riderval, rider))
        rideraux.sort()
        for r in rideraux:
            rno = r[2]
            rider = self.ridermap[rno]
            self.riders.append([
                rno, rider[u'name'], rider[u'club'], rider[u'cat'],
                rider[u'sex'],
                unicode(rider[u'points'])
            ])
            # also write to result thingo
            self.riderres.append([
                u'', rno, rider[u'namestr'], rider[u'sex'] + rider[u'cat'],
                u'',
                unicode(rider[u'points'])
            ])

        # creat aux sorting for team points
        bestteam = None
        bestavg = 0
        teamaux = []
        for team in self.teammap:
            pts = self.teammap[team][u'points']
            count = len(self.teammap[team][u'riders'])
            avg = float(pts) / count
            self.teammap[team][u'average'] = avg
            if avg > bestavg:
                bestteam = team
                bestavg = avg
            teamaux.append((-pts, -avg, team))
        teamaux.sort()
        for t in teamaux:
            name = t[2]
            team = self.teammap[name]
            pstr = unicode(team[u'points'])
            star = u''
            if name == bestteam:
                star = u'\u2605 '
            astr = '{0}{1:0.1f}'.format(star, team[u'average'])
            self.teamres.append([u'', u'', name, u'', astr, pstr])

        # update timerstat
        # if all results:
        #  self.timerstat = u'result'
        # if any results...
        #  self.timerstat = u'standing'
        # else:
        #  self.timerstat = u'idle'

        return

    def key_event(self, widget, event):
        """Race window key press handler."""
        if event.type == gtk.gdk.KEY_PRESS:
            key = gtk.gdk.keyval_name(event.keyval) or u'None'
            if event.state & gtk.gdk.CONTROL_MASK:
                if key == key_abort or key == key_reannounce:
                    # override ctrl+f5
                    self.recalculate()
                    glib.idle_add(self.delayed_announce)
                    return True
            elif key[0] == u'F':
                if key == key_startlist:
                    self.do_startlist()
                    glib.idle_add(self.delayed_announce)
                    return True
                elif key == key_results:
                    self.do_places()
                    glib.idle_add(self.delayed_announce)
                    return True
        return False

    def delayed_announce(self):
        """Initialise the announcer's screen after a delay."""
        ## TODO because # riders often exceeds 24 - requires paging
        return False
        if self.winopen:
            # clear page
            self.meet.txt_clear()
            self.meet.txt_title(u' '.join([
                u'Event', self.evno, u':', self.event[u'pref'],
                self.event[u'info']
            ]))
            self.meet.txt_line(1)
            self.meet.txt_line(19)

            # write out riders
            lmedal = ''
            posoft = 0
            l = 4
            for r in self.riders:
                if l > 17:
                    l = 4
                    posoft += 41
                plstr = u''
                pls = r[COL_PLACE].decode(u'utf-8')
                if pls:
                    plstr = pls
                    if plstr.isdigit():
                        plstr += u'.'
                plstr = strops.truncpad(plstr, 3, u'l')
                bibstr = strops.truncpad(r[COL_BIB].decode(u'utf-8'), 3, u'r')
                clubstr = u''
                tcs = r[COL_CLUB].decode(u'utf-8')
                if tcs and tcs <= 3:
                    clubstr = u' (' + tcs + u')'
                namestr = strops.truncpad(
                    strops.fitname(r[COL_FIRST].decode(u'utf-8'),
                                   r[COL_LAST].decode(u'utf-8'),
                                   25 - len(clubstr)) + clubstr, 25)
                medal = r[COL_MEDAL].decode(u'utf-8')
                if lmedal != u'' and medal == u'':
                    l += 1  # gap to medals
                lmedal = medal
                ol = [plstr, bibstr, namestr, medal]
                self.meet.txt_postxt(
                    l, posoft, u' '.join([plstr, bibstr, namestr, medal]))
                l += 1

        return False

    def do_startlist(self):
        """Show result on scoreboard."""
        # Don't show for agg
        return False

    def do_places(self):
        """Show race result on scoreboard."""
        # Don't show for agg
        return False
        ## todo: DISPLAY CLUB STANDINGS
        resvec = []
        count = 0
        teamnames = False
        name_w = self.meet.scb.linelen - 12
        fmt = [(3, u'l'), (4, u'r'), u' ', (name_w, u'l'), (4, u'r')]
        if self.series and self.series[0].lower() == u't':
            teamnames = True
            name_w = self.meet.scb.linelen - 8
            fmt = [(3, u'l'), u' ', (name_w, u'l'), (4, u'r')]

        for r in self.riders:
            plstr = r[COL_PLACE].decode('utf-8')
            if plstr.isdigit():
                plstr = plstr + u'.'
            no = r[COL_BIB]
            first = r[COL_FIRST].decode('utf-8')
            last = r[COL_LAST].decode('utf-8')
            club = r[COL_CLUB].decode('utf-8')
            if not teamnames:
                resvec.append(
                    [plstr, no,
                     strops.fitname(first, last, name_w), club])
            else:
                resvec.append([plstr, first, club])
            count += 1
        self.meet.scbwin = None
        header = self.meet.racenamecat(self.event)
        ## TODO: Flag Provisional
        evtstatus = u'Final Classification'.upper()
        self.meet.scbwin = scbwin.scbtable(scb=self.meet.scb,
                                           head=self.meet.racenamecat(
                                               self.event),
                                           subhead=evtstatus,
                                           coldesc=fmt,
                                           rows=resvec)
        self.meet.scbwin.reset()
        return False

    def shutdown(self, win=None, msg='Exiting'):
        """Terminate race object."""
        _log.debug('Shutdown event %s: %s', self.evno, msg)
        if not self.readonly:
            self.saveconfig()
        self.winopen = False

    def timercb(self, e):
        """Handle a timer event."""
        return False

    def timeout(self):
        """Update scoreboard and respond to timing events."""
        if not self.winopen:
            return False
        return True

    def do_properties(self):
        """Run race properties dialog."""
        prfile = os.path.join(metarace.UI_PATH, u'aggregate_properties.ui')
        _log.debug(u'Building event properties from %r', prfile)
        b = gtk.Builder()
        b.add_from_file(prfile)
        dlg = b.get_object(u'properties')
        dlg.set_transient_for(self.meet.window)
        se = b.get_object(u'race_qualifying_entry')
        se.set_text(self.qualifying)
        ee = b.get_object(u'race_events_entry')
        ee.set_text(self.events)
        pe = b.get_object(u'race_points_entry')
        pe.set_text(self.pointspec)
        me = b.get_object(u'race_double_entry')
        me.set_text(self.double)
        response = dlg.run()
        if response == 1:  # id 1 set in glade for "Apply"
            _log.debug(u'Updating event properties')
            self.pointspec = pe.get_text().decode(u'utf-8')
            self.double = me.get_text().decode(u'utf-8')
            self.events = ee.get_text().decode(u'utf-8')
            self.qualifying = se.get_text().decode(u'utf-8')
            self.recalculate()
            glib.idle_add(self.delayed_announce)
        else:
            _log.debug(u'Edit event properties cancelled')

        # if prefix is empty, grab input focus
        if not self.prefix_ent.get_text():
            self.prefix_ent.grab_focus()
        dlg.destroy()

    def destroy(self):
        """Signal race shutdown."""
        if self.context_menu is not None:
            self.context_menu.destroy()
        self.frame.destroy()

    def show(self):
        """Show race window."""
        self.frame.show()

    def hide(self):
        """Hide race window."""
        self.frame.hide()

    def update_expander_lbl_cb(self):
        """Update race info expander label."""
        self.info_expand.set_label(self.meet.infoline(self.event))

    def editent_cb(self, entry, col):
        """Shared event entry update callback."""
        if col == u'pref':
            self.event[u'pref'] = entry.get_text().decode(u'utf-8')
        elif col == u'info':
            self.event[u'info'] = entry.get_text().decode(u'utf-8')
        self.update_expander_lbl_cb()

    def editcol_db(self, cell, path, new_text, col):
        """Cell update with writeback to meet."""
        new_text = new_text.decode(u'utf-8', 'replace').strip()
        self.riders[path][col] = new_text
        glib.idle_add(self.meet.rider_edit,
                      self.riders[path][COL_BIB].decode(u'utf-8'), self.series,
                      col, new_text)

    def __init__(self, meet, event, ui=True):
        """Constructor."""
        self.meet = meet
        self.event = event  # Note: now a treerowref
        self.evno = event[u'evid']
        self.evtype = event[u'type']
        self.series = event[u'seri']
        self.configfile = meet.event_configfile(self.evno)
        _log.debug(u'Init event %s', self.evno)

        # race run time attributes
        self.onestart = True  # always true for autospec classification
        self.readonly = not ui
        self.winopen = ui
        self.comments = []
        self.events = u''
        self.qualifying = u''
        self.pointspec = u''
        self.double = u''
        self.timerstat = u'idle'  # fake a timerstat and use for result stat
        self.ridermap = {}
        self.teammap = {}
        self.teamres = []
        self.teamavgres = []
        self.afterev = None

        self.riders = gtk.ListStore(
            gobject.TYPE_STRING,  # 0 bib
            gobject.TYPE_STRING,  # 1 name
            gobject.TYPE_STRING,  # 2 club
            gobject.TYPE_STRING,  # 3 cat
            gobject.TYPE_STRING,  # 4 sex
            gobject.TYPE_STRING)  # 5 points

        uifile = os.path.join(metarace.UI_PATH, u'classification.ui')
        _log.debug(u'Building event interface from %r', uifile)
        b = gtk.Builder()
        b.add_from_file(uifile)

        self.frame = b.get_object(u'classification_vbox')
        self.frame.connect(u'destroy', self.shutdown)

        # info pane
        self.info_expand = b.get_object(u'info_expand')
        b.get_object(u'classification_info_evno').set_text(self.evno)
        self.showev = b.get_object(u'classification_info_evno_show')
        self.prefix_ent = b.get_object(u'classification_info_prefix')
        self.prefix_ent.set_text(self.event[u'pref'])
        self.prefix_ent.connect(u'changed', self.editent_cb, u'pref')
        self.info_ent = b.get_object(u'classification_info_title')
        self.info_ent.set_text(self.event[u'info'])
        self.info_ent.connect(u'changed', self.editent_cb, u'info')

        self.context_menu = None
        if ui:
            # riders pane
            t = gtk.TreeView(self.riders)
            self.view = t
            t.set_rules_hint(True)

            # riders columns
            uiutil.mkviewcoltxt(t, u'No.', COL_BIB, calign=1.0)
            uiutil.mkviewcoltxt(t, u'Name', COL_NAME, expand=True)
            uiutil.mkviewcoltxt(t, u'Club', COL_CLUB)
            uiutil.mkviewcoltxt(t, u'Cat', COL_CAT)
            uiutil.mkviewcoltxt(t, u'Sex', COL_SEX)
            uiutil.mkviewcoltxt(t, u'Points', COL_POINTS, calign=1.0)
            t.show()
            b.get_object(u'classification_result_win').add(t)
            b.connect_signals(self)



"""Countback support."""


class countback(object):
    """Wrapper for countback store/compare."""
    __hash__ = None

    def __init__(self, cbstr=None):
        self.__store = {}
        if cbstr is not None:
            self.fromstring(cbstr)

    def maxplace(self):
        """Return maximum non-zero place."""
        ret = 0
        if len(self.__store) > 0:
            ret = max(self.__store.keys())
        return ret

    def fromstring(self, cbstr):
        propmap = {}
        cbvec = cbstr.split(u',')
        if len(cbvec) > 0:
            for i in range(0, len(cbvec)):
                if cbvec[i].isdigit():
                    propmap[i] = int(cbvec[i])
        self.__store = {}
        for k in propmap:
            self.__store[k] = propmap[k]

    def __repr__(self):
        return 'countback({})'.format(repr(self.__unicode__()))

    def __str__(self):
        return str(self.__unicode__())

    def __unicode__(self):
        ret = []
        for i in range(0, self.maxplace() + 1):
            if i in self.__store and self.__store[i] != 0:
                ret.append(unicode(self.__store[i]))
            else:
                ret.append(u'-')
        return u','.join(ret)

    def __len__(self):
        return len(self.__store.len)

    def __getitem__(self, key):
        if key in self.__store:
            return self.__store[key]
        else:
            return 0

    def __setitem__(self, key, value):
        if isinstance(key, int):
            self.__store[key] = value
        else:
            raise AttributeError(u'Countback keys must be integer')

    def __delitem__(self, key):
        del (self.__store[key])

    def __iter__(self):
        return self.__store.iterkeys()

    def iterkeys(self):
        return self.__store.iterkeys()

    def __contains__(self, item):
        return item in self.__store

    def __lt__(self, other):
        if not isinstance(other, countback):
            return NotImplemented
        ret = False  # assume all same
        for i in range(0, max(self.maxplace(), other.maxplace()) + 1):
            a = self[i]
            b = other[i]
            if a != b:
                ret = a < b
                break
        return ret

    def __le__(self, other):
        if not isinstance(other, countback):
            return NotImplemented
        ret = True  # assume all same
        for i in range(0, max(self.maxplace(), other.maxplace()) + 1):
            a = self[i]
            b = other[i]
            if a != b:
                ret = a < b
                break
        return ret

    def __eq__(self, other):
        if not isinstance(other, countback):
            return NotImplemented
        ret = True
        for i in range(0, max(self.maxplace(), other.maxplace()) + 1):
            if self[i] != other[i]:
                ret = False
                break
        return ret

    def __ne__(self, other):
        if not isinstance(other, countback):
            return NotImplemented
        ret = False
        for i in range(0, max(self.maxplace(), other.maxplace()) + 1):
            if self[i] != other[i]:
                ret = True
                break
        return ret

    def __gt__(self, other):
        if not isinstance(other, countback):
            return NotImplemented
        ret = False  # assume all same
        for i in range(0, max(self.maxplace(), other.maxplace()) + 1):
            a = self[i]
            b = other[i]
            if a != b:
                ret = a > b
                break
        return ret

    def __ge__(self, other):
        if not isinstance(other, countback):
            return NotImplemented
        ret = True  # assume all same
        for i in range(0, max(self.maxplace(), other.maxplace()) + 1):
            a = self[i]
            b = other[i]
            if a != b:
                ret = a > b
                break
        return ret

    def __add__(self, other):
        """Add two countbacks together and return a new cb>=self >=other."""
        if not isinstance(other, countback):
            return NotImplemented
        ret = countback(str(self))
        for i in range(0, max(self.maxplace(), other.maxplace()) + 1):
            ret[i] += other[i]
        return ret



# SPDX-License-Identifier: MIT
"""Tag Heuer/Chronelec Decoder Interface."""

# todo: handle man, box, and cel combinations:
#	CELL1   <STA 000255 00:00'31"957 01 01 1 1555>
#	CELL2   <BOX 000255 00:01'32"663 01 01 1 1552>
#	MAN     <MAN 000000 00:18'16"964 01 01 1 1541>
#	box L   <B01 093367 03:45'08"388 44 01 0 1522>
#	box R   <B02 093367 03:45'53"976 74 01 0 1529>
#	box Cel <B01 000255 03:47'07"566 00 01 0 1497>

# For connections to multiple decoders, use thbchub

import Queue
import logging
import serial
import socket
import time

from . import (decoder, DECODER_LOG_LEVEL)
from metarace import sysconf
from metarace import tod
from libscrc import mcrf4xx

_log = logging.getLogger(u'metarace.decoder.thbc')
_log.setLevel(logging.DEBUG)

THBC_BAUD = 19200
THBC_UDP_PORT = 2008
THBC_ENCODING = u'iso8859-1'

# THbC protocol messages
ESCAPE = b'\x1b'
HELOCMD = b'MR1'
STOPCMD = ESCAPE + b'\x13\x5c'
REPEATCMD = ESCAPE + b'\x12'
ACKCMD = ESCAPE + b'\x11'

STATCMD = ESCAPE + b'\x05'  # fetch status
CHKCMD = ESCAPE + b'\x06'  # UNKNOWN
STARTCMD = ESCAPE + b'\x07'  # start decoder
SETCMD = ESCAPE + b'\x08'  # set configuration
IPCMD = ESCAPE + b'\x09'  # set IP configuration
QUECMD = ESCAPE + b'\x10'  # fetch configuration

STALVL = ESCAPE + b'\x1e'
BOXLVL = ESCAPE + b'\x1f'

NACK = b'\x07'
CR = b'\x0d'
LF = b'\x0a'
SETTIME = ESCAPE + b'\x48'
STATSTART = b'['
PASSSTART = b'<'

# decoder config consts
IPCONFIG_LEN = 16
CONFIG_LEN = 27
CONFIG_TOD = 0
CONFIG_GPS = 1
CONFIG_TZ_HOUR = 2
CONFIG_TZ_MIN = 3
CONFIG_485 = 4
CONFIG_FIBRE = 5
CONFIG_PRINT = 6
CONFIG_MAX = 7
CONFIG_PROT = 8
CONFIG_PULSE = 9
CONFIG_PULSEINT = 10
CONFIG_CELLSYNC = 11
CONFIG_CELLTOD_HOUR = 12
CONFIG_CELLTOD_MIN = 13
CONFIG_TONE_STA = 15
CONFIG_TONE_BOX = 17
CONFIG_TONE_MAN = 19
CONFIG_TONE_CEL = 21
CONFIG_TONE_BXX = 23
CONFIG_ACTIVE_LOOP = 14
CONFIG_SPARE = 25
CONFIG_FLAGS = {
    CONFIG_TOD: u'Time of Day',
    CONFIG_GPS: u'GPS Sync',
    CONFIG_TZ_HOUR: u'Timezone Hour',
    CONFIG_TZ_MIN: u'Timezone Min',
    CONFIG_485: u'Distant rs485',
    CONFIG_FIBRE: u'Distant Fibre',
    CONFIG_PRINT: u'Serial Print',
    CONFIG_MAX: u'Detect Max',
    CONFIG_PROT: u'Protocol',
    CONFIG_PULSE: u'Sync Pulse',
    CONFIG_PULSEINT: u'Sync Interval',
    CONFIG_CELLSYNC: u'CELL Sync',
    CONFIG_CELLTOD_HOUR: u'CELL Sync Hour',
    CONFIG_CELLTOD_MIN: u'CELL Sync Min',
    CONFIG_TONE_STA: u'STA Tone',
    CONFIG_TONE_BOX: u'BOX Tone',
    CONFIG_TONE_MAN: u'MAN Tone',
    CONFIG_TONE_CEL: u'CEL Tone',
    CONFIG_TONE_BXX: u'BXX Tone',
    CONFIG_ACTIVE_LOOP: u'Active Loop',
    CONFIG_SPARE: u'[spare]'
}
DEFAULT_IPCFG = {
    u'IP': u'192.168.0.10',
    u'Netmask': u'255.255.255.0',
    u'Gateway': u'0.0.0.0',
    u'Host': u'192.168.0.255'
}
DEFPORT = u'/dev/ttyS0'


def thbc_sum(msgstr=b''):
    """Return sum of character values as decimal string."""
    ret = 0
    for ch in msgstr:
        ret = ret + ord(ch)
    return '{0:04d}'.format(ret)


def val2hexval(val):
    """Convert int to decimal digit equivalent hex byte."""
    ret = 0x00
    ret |= ((val // 10) & 0x0f) << 4  # msd	97 -> 0x90
    ret |= (val % 10) & 0x0f  # lsd   97 -> 0x07
    return ret


def hexval2val(hexval):
    """Unconvert a decimal digit equivalent hex byte to int."""
    ret = 10 * (hexval >> 4)  # tens 0x97 -> 90
    ret += hexval & 0x0f  # ones 0x97 ->  7
    return ret


class thbc(decoder):
    """Tag Heuer / Chronelec thread object class."""

    def __init__(self):
        decoder.__init__(self)
        self._boxname = None
        self._version = u''
        self._decoderconfig = {}
        self._decoderipconfig = {}
        self._io = None
        self._cksumerr = 0
        self._rdbuf = b''  # bytestring read buffer

    # API overrides
    def status(self):
        """Request status message from decoder."""
        self.write(STATCMD)

    def connected(self):
        return self._io is not None and self._boxname is not None

    def stop_session(self):
        """Send a stop command to decoder."""
        self.write(STOPCMD)

    def start_session(self):
        """Send a depart command to decoder."""
        self.write(STARTCMD)

    def clear(self):
        """Start a new session and request time sync."""
        self.stop_session()
        self.start_session()
        self.sync()

    def get_config(self):
        """Request decoder configuration."""
        self.write(QUECMD)

    def ipconfig(self):
        """Request sanity check in decoder thread."""
        self._cqueue.put_nowait((u'_ipcfg', None))

    # Device-specific functions
    def _close(self):
        if self._io is not None:
            _log.debug(u'Close connection')
            cp = self._io
            self._io = None
            try:
                cp.close()
            except Exception as e:
                _log.debug(u'%s closing io: %s', e.__class__.__name__, e)

    def _port(self, port):
        """Re-establish connection to supplied device port."""
        self._close()
        s = None
        self._rdbuf = b''
        if u'/' not in port and u'.' in port:
            _log.debug(u'Attempting UDP on %r', port)
            s = dgram(port, THBC_UDP_PORT)
        else:
            # assume device file
            s = serial.Serial(port, THBC_BAUD, rtscts=False, timeout=0.2)
        self._boxname = None
        self._io = s
        self._write(QUECMD)
        # queue sane through command loop
        time.sleep(0.2)
        self.sane()

    def _sync(self, data=None):
        _log.debug(u'Performing blocking sync')
        acceptval = tod.tod(u'0.001')
        nt = tod.now()
        diff = nt - nt.truncate(0)
        while diff > acceptval and diff < tod.ONE:
            time.sleep(0.0005)
            nt = tod.now()
            diff = nt - nt.truncate(0)
        self._write(self._set_time_cmd(nt))
        _log.info(u'Set decoder time: %r', nt.meridiem())

    def _ipcfg(self, data=None):
        """Alter the attached decoder's IP address."""
        ipcfg = sysconf.get(u'thbc', u'ipconfig')
        cmd = b'\x09\x09'
        for i in [u'IP', u'Netmask', u'Gateway', u'Host']:
            if i not in ipcfg:
                ipcfg[i] = DEFAULT_IPCFG[i]
            cmd += socket.inet_aton(socket.gethostbyname(ipcfg[i]))
        _log.info(u'Attempting IP config update')
        self._v3_cmd(cmd)

    def _sane(self, data=None):
        """Check decoder config against system settings."""
        doconf = False
        if self._boxname is not None:
            if sysconf.has_option(u'thbc', u'decoderconfig'):
                oconf = sysconf.get(u'thbc', u'decoderconfig')
                for flag in self._decoderconfig:
                    key = CONFIG_FLAGS[flag]
                    if key in oconf:
                        if oconf[key] != self._decoderconfig[flag]:
                            _log.info(u'Key mismatch: %r', key)
                            self._decoderconfig[flag] = oconf[key]
                            doconf = True
        else:
            _log.info(u'Decoder not connected')

        # re-write config if required
        if doconf:
            _log.info(u'Re-configuring %r', self._boxname)
            self._set_config()

        # force decoder levels
        if sysconf.has_option(u'thbc', u'levels'):
            lvl = sysconf.get(u'thbc', u'levels')
            self._setlvl(box=lvl[0], sta=lvl[1])

    def _v3_cmd(self, cmdstr=b''):
        """Pack and send a v3 command directly to port."""
        crc = mcrf4xx(cmdstr)
        crcstr = chr(crc >> 8) + chr(crc & 0xff)  # Py2
        self._write(ESCAPE + cmdstr + crcstr + b'>')

    def _serialise_config(self):
        """Convert current decoder setting into a config string"""
        obuf = bytearray(CONFIG_LEN)
        # fill in level bytes
        obuf[CONFIG_SPARE] = 0x13  # will be fixed by subsequent levelset
        obuf[CONFIG_SPARE + 1] = 0x15

        # fill in tone values
        for opt in [
                CONFIG_TONE_STA, CONFIG_TONE_BOX, CONFIG_TONE_MAN,
                CONFIG_TONE_CEL, CONFIG_TONE_BXX
        ]:
            if opt in self._decoderconfig:
                obuf[opt] = val2hexval(self._decoderconfig[opt] // 100)  # xx00
                obuf[opt + 1] = val2hexval(self._decoderconfig[opt] %
                                           100)  # 00xx

        # fill in single byte values
        for opt in [
                CONFIG_TZ_HOUR, CONFIG_TZ_MIN, CONFIG_PROT, CONFIG_PULSEINT,
                CONFIG_CELLTOD_HOUR, CONFIG_CELLTOD_MIN
        ]:
            if opt in self._decoderconfig:
                obuf[opt] = val2hexval(self._decoderconfig[opt] % 100)  # ??

        # fill in flags
        for opt in [
                CONFIG_TOD, CONFIG_GPS, CONFIG_485, CONFIG_FIBRE, CONFIG_PRINT,
                CONFIG_MAX, CONFIG_PULSE, CONFIG_CELLSYNC, CONFIG_ACTIVE_LOOP
        ]:
            if opt in self._decoderconfig:
                if self._decoderconfig[opt]:
                    obuf[opt] = 0x01
        return bytes(obuf)

    def _set_config(self):
        """Write current configuration to decoder."""
        cmd = b'\x08\x08' + self._serialise_config()
        self._v3_cmd(cmd)
        self._write(QUECMD)

    def _set_date(self, timestruct=None):
        """Set the date on the decoder."""
        if timestruct is None:
            timestruct = time.localtime()
        _log.debug(u'Set date on decoder: %s',
                   time.strftime('%Y-%m-%d', timestruct))
        cmd = bytearray(5)
        cmd[0] = 0x0a
        cmd[1] = 0x0a
        cmd[2] = 0xff & timestruct[2]  # day
        cmd[3] = 0xff & timestruct[1]  # month
        cmd[4] = 0xff & (timestruct[0] - 2000)  # year, after 2000
        self._v3_cmd(bytes(cmd))

    def _setlvl(self, box=u'10', sta=u'10'):
        """Set the read level on box and sta channels."""
        # TODO: verify opts
        self.write(BOXLVL + box.encode(THBC_ENCODING))
        self.write(STALVL + sta.encode(THBC_ENCODING))

    def _set_time_cmd(self, t):
        """Return a set time command string for the provided time of day."""
        body = bytearray(4)
        s = int(t.timeval)
        body[0] = s // 3600  # hours
        body[1] = (s // 60) % 60  # minutes
        body[2] = s % 60  # seconds
        body[3] = 0x74
        return SETTIME + bytes(body)

    def _parse_config(self, msg):
        # decoder configuration message.
        ibuf = bytearray(msg)
        self._decoderconfig = {}
        for flag in sorted(CONFIG_FLAGS):  # import all
            # tone values
            if flag in [
                    CONFIG_TONE_STA, CONFIG_TONE_BOX, CONFIG_TONE_MAN,
                    CONFIG_TONE_CEL, CONFIG_TONE_BXX
            ]:
                self._decoderconfig[flag] = 100 * hexval2val(ibuf[flag])
                self._decoderconfig[flag] += hexval2val(ibuf[flag + 1])

            # single byte values
            elif flag in [
                    CONFIG_TZ_HOUR, CONFIG_TZ_MIN, CONFIG_PROT,
                    CONFIG_PULSEINT, CONFIG_CELLTOD_HOUR, CONFIG_CELLTOD_MIN
            ]:
                self._decoderconfig[flag] = hexval2val(ibuf[flag])

            # 'booleans'
            elif flag in [
                    CONFIG_TOD, CONFIG_GPS, CONFIG_485, CONFIG_FIBRE,
                    CONFIG_PRINT, CONFIG_MAX, CONFIG_PULSE, CONFIG_CELLSYNC,
                    CONFIG_ACTIVE_LOOP
            ]:
                self._decoderconfig[flag] = bool(ibuf[flag])

        self._boxname = u''
        for c in msg[43:47]:
            self._boxname += unichr(ord(c) + ord('0'))
        self._version = unicode(hexval2val(ibuf[47]))
        stalvl = hex(ord(msg[25]))  # ? question this
        boxlvl = hex(ord(msg[26]))
        _log.info(u'Info Decoder ID: %s', self._boxname)
        _log.debug(u'Info Firmware Version: %r', self._version)
        _log.debug(u'Levels: STA=%r, BOX=%r', stalvl, boxlvl)
        self._decoderipconfig[u'IP'] = socket.inet_ntoa(msg[27:31])
        self._decoderipconfig[u'Mask'] = socket.inet_ntoa(msg[31:35])
        self._decoderipconfig[u'Gateway'] = socket.inet_ntoa(msg[35:39])
        self._decoderipconfig[u'Host'] = socket.inet_ntoa(msg[39:43])
        for key in [u'IP', u'Mask', u'Gateway', u'Host']:
            _log.debug(u'%r: %r', key, self._decoderipconfig[key])

    def _parse_message(self, msg, ack=True):
        """Return tod object from timing msg or None."""
        ret = None
        if len(msg) > 4:
            if msg[0] == PASSSTART:  # RFID message
                idx = msg.find(b'>')
                if idx == 37:  # Valid length
                    data = msg[1:33]
                    msum = msg[33:37]
                    tsum = thbc_sum(data)
                    if tsum == msum:  # Valid 'sum'
                        pvec = data.decode(THBC_ENCODING).split()
                        istr = pvec[3] + u':' + pvec[5]
                        rstr = pvec[1].lstrip(u'0')
                        cstr = u'C1'
                        if pvec[0] == u'BOX':
                            cstr = u'C2'
                        elif pvec[0] == u'MAN':
                            cstr = u'C0'
                        if pvec[5] == u'3':  # LOW BATTERY ALERT
                            _log.warning(u'Low battery on %r', rstr)
                        ret = tod.tod(pvec[2],
                                      index=istr,
                                      chan=cstr,
                                      refid=rstr,
                                      source=self._boxname)
                        # Log a hardware-specific passing
                        _log.log(DECODER_LOG_LEVEL, msg.strip())
                        if ack:
                            self._write(ACKCMD)  # Acknowledge if ok
                        self._cksumerr = 0
                    else:
                        _log.warning(u'Invalid checksum: %r != %r: %r', tsum,
                                     msum, msg)
                        self._cksumerr += 1
                        if self._cksumerr > 3:
                            # assume error on decoder, so acknowledge and
                            # continue with log
                            # NOTE: This path is triggered when serial comms
                            # fail and a tag read happens before a manual trig
                            _log.error(u'Erroneous message from decoder')
                            if ack:
                                self._write(ACKCMD)
                else:
                    _log.debug(u'Invalid message: %r', msg)
            elif msg[0] == STATSTART:  # Status message
                data = msg[1:22]
                pvec = data.decode(THBC_ENCODING).split()
                if len(pvec) == 5:
                    _log.info(u'%r@%s Noise:%s/%s Levels:%s/%s', self._boxname,
                              pvec[0], pvec[1], pvec[2], pvec[3], pvec[4])
                else:
                    _log.info(u'Invalid status: %r', msg)
            elif b'+++' == msg[0:3] and len(msg) > 53:
                self._parse_config(msg[3:])
            else:
                pass
        else:
            _log.debug(u'Short message: %r', msg)
        return ret

    def _ipcompletion(self):
        """Blocking wait for ipconfig completion - horrible."""
        _log.info(u'IP Config')
        time.sleep(10)
        self.write(QUECMD)

    def _read(self):
        """Read messages from the decoder until a timeout condition."""
        ch = self._io.read(1)
        while ch != b'':
            if ch == LF and len(self._rdbuf) > 0 and self._rdbuf[-1] == CR:
                # Return ends the current 'message', if preceeded by return
                self._rdbuf += ch  # include trailing newline
                #_log.debug(u'RECV: %r', self._rdbuf)
                t = self._parse_message(self._rdbuf.lstrip(b'\0'))
                if t is not None:
                    self._trig(t)
                self._rdbuf = b''
            elif len(self._rdbuf) > 40 and b'\x1e\x86\x98' in self._rdbuf:
                # Assume acknowledge from IP Command
                #_log.debug(u'RECV: %r', self._rdbuf)
                self._rdbuf = b''
                self._ipcompletion()
            else:
                self._rdbuf += ch
            ch = self._io.read(1)

    def _write(self, msg):
        if self._io is not None:
            self._io.write(msg)
            #_log.debug(u'SEND: %r', msg)

    def run(self):
        """Decoder main loop."""
        _log.debug(u'Starting')
        self._running = True
        while self._running:
            try:
                c = None
                if self._io is not None:
                    # Read responses until response complete or timeout
                    try:
                        self._read()
                    except socket.timeout:
                        pass
                    c = self._cqueue.get_nowait()
                else:
                    c = self._cqueue.get()
                self._cqueue.task_done()
                self._proccmd(c)
            except Queue.Empty:
                pass
            except (serial.SerialException, socket.error) as e:
                self._close()
                self._boxname = None
                _log.error(u'%s: %s', e.__class__.__name__, e)
            except Exception as e:
                _log.critical(u'%s: %s', e.__class__.__name__, e)
                self._boxname = None
                self._running = False
        self.setcb()
        _log.debug(u'Exiting')


class dgram(object):
    """Serial-like UDP port object."""

    def __init__(self, host, port):
        self._host = host
        self._port = port
        self._s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self._s.settimeout(0.2)
        self._s.bind((u'', self._port))
        self._buf = b''

    def read(self, sz=1):
        ret = b''  # check this condition
        if len(self._buf) == 0:
            nb, addr = self._s.recvfrom(4096)  # timeout raises exception
            if addr[0] == self._host:
                self._buf += nb
        if len(self._buf) > 0:
            ret = self._buf[0]
            self._buf = self._buf[1:]
        return ret

    def write(self, buf=b''):
        return self._s.sendto(buf, (self._host, self._port))

    def close(self):
        self._s.close()



"""Transitional Python 2.x unicode csv helper module."""

# Copied from the Python manual:
#
#  https://docs.python.org/release/2.7.16/library/csv.html#csv-examples
#
# Note: - Contrary to the typical behaviour of a utf-8 decoder,
#         invalid, erroeneous or non-unicode chars are silently ignored
# 	- Some auto-generated CSV with newlines in cells will not read properly
#	- This module _has_ to go.

import csv, codecs, cStringIO

QUOTE_MINIMAL = csv.QUOTE_MINIMAL


class UTF8Recoder:
    """
    Iterator that reads an encoded stream and reencodes the input to UTF-8
    """

    def __init__(self, f, encoding):
        self.reader = codecs.getreader(encoding)(f, errors='ignore')

    def __iter__(self):
        return self

    def next(self):
        return self.reader.next().encode('utf-8', 'ignore')


class UnicodeReader:
    """
    A CSV reader which will iterate over lines in the CSV file "f",
    which is encoded in the given encoding.
    """

    def __init__(self, f, dialect=None, encoding="utf-8", **kwds):
        if dialect is None:
            try:
                dialect = csv.Sniffer().sniff(f.read(1024))
            except:
                pass
            f.seek(0)
        f = UTF8Recoder(f, encoding)
        self.reader = csv.reader(f, dialect=dialect, **kwds)

    def next(self):
        row = self.reader.next()
        return [unicode(s, "utf-8") for s in row]

    def __iter__(self):
        return self


class UnicodeWriter:
    """
    A CSV writer which will write rows to CSV file "f",
    which is encoded in the given encoding.
    """

    def __init__(self,
                 f,
                 dialect=csv.excel,
                 encoding="utf-8",
                 quoting=csv.QUOTE_NONNUMERIC,
                 **kwds):
        # Redirect output to a queue
        dialect.quoting = quoting
        self.queue = cStringIO.StringIO()
        self.writer = csv.writer(self.queue, dialect=dialect, **kwds)
        self.stream = f
        self.encoder = codecs.getincrementalencoder(encoding)(errors='ignore')

    def writerow(self, row):
        self.writer.writerow([s.encode("utf-8", 'ignore') for s in row])
        # Fetch UTF-8 output from the queue ...
        data = self.queue.getvalue()
        data = data.decode("utf-8", 'ignore')
        # ... and reencode it into the target encoding
        data = self.encoder.encode(data)
        # write to the target stream
        self.stream.write(data)
        # empty queue
        self.queue.truncate(0)

    def writerows(self, rows):
        for row in rows:
            self.writerow(row)



"""export result files in a thread"""

import threading
import subprocess
import logging
import os

import metarace

MIRROR_CMD = u'echo'  # Command/Argument defaults
MIRROR_ARGS = [u'dummy', u'srcdir={srcdir}', u'dstdir={dstdir}']
_log = logging.getLogger(u'metarace.export')
_log.setLevel(logging.DEBUG)


class mirror(threading.Thread):
    """Mirror thread object class."""

    def __init__(self,
                 callback=None,
                 callbackdata=None,
                 localpath=u'.',
                 remotepath=None,
                 mirrorcmd=None,
                 arguments=None,
                 data=None):
        """Construct mirror thread object."""
        threading.Thread.__init__(self)
        self.daemon = True
        self.__cb = None
        if callback is not None:
            self.__cb = callback
        self.__cbdata = None
        if callbackdata is not None:
            self.__cbdata = callbackdata
        self.__localpath = localpath
        self.__remotepath = u''
        if remotepath is not None:
            self.__remotepath = remotepath

        # config starts with module defaults
        self.__mirrorcmd = MIRROR_CMD
        self.__arguments = MIRROR_ARGS

        # then overwrite from sysconf - if present
        if metarace.sysconf.has_section(u'export'):
            if metarace.sysconf.has_option(u'export', u'command'):
                self.__mirrorcmd = metarace.sysconf.get(u'export', u'command')
            if metarace.sysconf.has_option(u'export', u'arguments'):
                self.__arguments = metarace.sysconf.get(
                    u'export', u'arguments')

        # and then finally allow override in object creation
        if mirrorcmd:
            self.__mirrorcmd = mirrorcmd
        if arguments:
            self.__arguments = arguments

        self.__data = data

    def set_remotepath(self, pathstr):
        """Set or clear the remote path value."""
        self.__remotepath = pathstr

    def set_localpath(self, pathstr):
        """Set or clear the local path value."""
        self.__localpath = pathstr

    def set_cb(self, func=None, cbdata=None):
        """Set or clear the event callback."""
        # if func is not callable, gtk mainloop will catch the error
        if func is not None:
            self.__cb = func
            self.__cbdata = cbdata
        else:
            self.__cb = None
            self.__cbdata = None

    def run(self):
        """Called via threading.Thread.start()."""
        running = True
        _log.debug(u'Starting')
        ret = None
        try:
            # format errors in arguments caught as exception
            arglist = [
                a.format(srcdir=self.__localpath,
                         dstdir=self.__remotepath,
                         command=self.__mirrorcmd,
                         data=self.__data) for a in self.__arguments
            ]
            arglist.insert(0, self.__mirrorcmd)

            _log.debug(u'Calling subprocess: %r', arglist)
            # TODO: convert to run w/timeout and I/O capture (py3)
            ret = subprocess.check_call(arglist)
            if self.__cb:
                self.__cb(ret, self.__cbdata)
        except Exception as e:
            _log.error(u'%s: %s', e.__class__.__name__, e)
        _log.info(u'Complete: Returned %r', ret)
